#include "GUIVerifyPanel.h"
#include "ui_GUIVerifyPanel.h"
#include "GUIMainWindow.h"
#include "GUICommon.h"
#include "UIApp.h"
#include "Matrix.h"

using namespace std;

GUIVerifyPanel::GUIVerifyPanel(QWidget *parent) :
	QDialog(parent), ui(new Ui::GUIVerifyPanel) {
	ui->setupUi(this);

	::setWindowTitle(this);

	// ファイルタイプ
	QComboBox *combo = ui->file_type;
	removeComboAll(combo);
	combo->addItem("(Undefined)", QVariant(HT::UIApp::UNDEFINEDID));
	combo->addItem("Separation-TF", QVariant(HT::UIApp::GSS));
	combo->addItem("Separation-Matrix", QVariant(HT::UIApp::GSSMAT));
	combo->addItem("Localization-TF", QVariant(HT::UIApp::MUSIC));

	// ファイル
	removeComboAll(ui->file1);
	removeComboAll(ui->file2);
	ui->file1->addItem("(Undefined)");
	ui->file2->addItem("(Undefined)");

	connectTargetChanged(true);
	targetChanged();
	adjustSize();
}

void GUIVerifyPanel::connectTargetChanged(bool value) {

	if (value) {
		connect(ui->file_type, SIGNAL(currentIndexChanged(int)), this, SLOT(targetChanged()));
		connect(ui->file1, SIGNAL(currentIndexChanged(int)), this, SLOT(targetChanged()));
		connect(ui->file2, SIGNAL(currentIndexChanged(int)), this, SLOT(targetChanged()));
	} else {
		disconnect(ui->file_type, SIGNAL(currentIndexChanged(int)), this, SLOT(targetChanged()));
		disconnect(ui->file1, SIGNAL(currentIndexChanged(int)), this, SLOT(targetChanged()));
		disconnect(ui->file2, SIGNAL(currentIndexChanged(int)), this, SLOT(targetChanged()));
	}
}

void GUIVerifyPanel::execCompare() {

	try {
		QWidget* widget1 = ui->file1->itemData(ui->file1->currentIndex()).value<QWidget*> ();
		QWidget* widget2 = ui->file1->itemData(ui->file2->currentIndex()).value<QWidget*> ();

		GUITabDocument* guidoc1 = dynamic_cast<GUITabDocument*> (widget1);
		HT::UIDocument* doc1 = guidoc1->getDoc();

		GUITabDocument* guidoc2 = dynamic_cast<GUITabDocument*> (widget2);
		HT::UIDocument* doc2 = guidoc2->getDoc();

		HT::UIData *data = doc1->getBaseUIData();
		HT::Transform transform1;
		TFMatrix<HT_COMPLEX_DATA_TYPE> *matrix1 = &transform1.commonTFMatrix(*data);

		HT::UIData *data2 = doc2->getBaseUIData();
		HT::Transform transform2;
		TFMatrix<HT_COMPLEX_DATA_TYPE> *matrix2 = &transform2.commonTFMatrix(*data2);

		dispComparedData(matrix1, matrix2);
	} catch (std::exception& exp) {
		DISP_DIALOG(exp);
		return;
	}
}

void GUIVerifyPanel::dispComparedData(TFMatrix<HT_COMPLEX_DATA_TYPE> *matrix1, TFMatrix<HT_COMPLEX_DATA_TYPE>* matrix2) {

	int tab1_error = 0;
	int tab2_error = 0;

	//マトリックスのサイズチェック
	if (matrix1->size().size() == matrix2->size().size()) {
		for (uint i = 0; i < matrix1->size().size(); i++) {
			if (matrix1->size()[i] != matrix2->size()[i]) {
				THROW_WARNING(tr("Incompatible matrix size", "比較対象のマトリックスのサイズが異なります。"));
			}
		}
	}

	vector<int> size1 = matrix1->size();
	vector<double> diff_temp;
	vector<double> diff_ratio_temp;
	double diff; // 差分
	double diff_ratio; // 差分率

	for (int i = 0; i < size1[0]; i++) {
		for (int j = 0; j < size1[1]; j++) {
			for (int k = 0; k < size1[2]; k++) {
				// データチェック
				if (!isfinite(matrix1->matrix[i][j][k].real())) {
					tab1_error++;
				}
				if (!isfinite(matrix1->matrix[i][j][k].imag())) {
					tab1_error++;
				}
				if (!isfinite(matrix2->matrix[i][j][k].real())) {
					tab2_error++;
				}
				if (!isfinite(matrix2->matrix[i][j][k].imag())) {
					tab2_error++;
				}
				// Amplitude の差分
				double re1 = matrix1->matrix[i][j][k].real();
				double im1 = matrix1->matrix[i][j][k].imag();
				double amp1 = sqrt(re1*re1 + im1*im1);

				double re2 = matrix2->matrix[i][j][k].real();
				double im2 = matrix2->matrix[i][j][k].imag();
				double amp2 = sqrt(re2*re2 + im2*im2);

				diff = amp1 - amp2;
				diff_ratio = diff / amp2;

				diff_temp.push_back(diff);
				diff_ratio_temp.push_back(diff_ratio);
			}
		}
	}
	// 差分の統計情報計算
	CheckVector<double> check_diff(diff_temp);
	check_diff.execCheck();
//	check_diff.print();

	CheckVector<double> check_diff_ratio(diff_ratio_temp);
	check_diff_ratio.execCheck();
//	check_diff_ratio.print();

	// 詳細の表示
	ui->result_box->show();

	ui->nan->setText(QString("%1").arg(check_diff.error));

	ui->min_diff->setText(QString("%1").arg(check_diff.min, 7, 'f', 4));
	ui->max_diff->setText(QString("%1").arg(check_diff.max, 7, 'f', 4));
	ui->avg_diff->setText(QString("%1").arg(check_diff.avg, 7, 'f', 4));

	ui->min_ratio->setText(QString("%1").arg(check_diff_ratio.min, 7, 'f', 4));
	ui->max_ratio->setText(QString("%1").arg(check_diff_ratio.max, 7, 'f', 4));
	ui->avg_ratio->setText(QString("%1").arg(check_diff_ratio.avg, 7, 'f', 4));

	// コンボボックス作成
	removeComboAll(ui->mic_combo);
	for (uint i = 0; i < matrix1->matrix[0].size(); i++) { // マイク数
		ui->mic_combo->addItem(QString("%1").arg(i), QVariant(i));
	}
	return;
}

void GUIVerifyPanel::dispComparedGraph(TFMatrix<HT_COMPLEX_DATA_TYPE> *matrix1, TFMatrix<HT_COMPLEX_DATA_TYPE>* matrix2, HT::UIData *data) {

	matrix2 = matrix1;
	// コンボボックスで選択したマイク番号
	int mic = ui->mic_combo->itemData(ui->mic_combo->currentIndex()).toInt();

	TFMatrix<HT_COMPLEX_DATA_TYPE> matrix1_copy = matrix1->range(HT::MUSIC::UIData::DM3, HT::MUSIC::UIData::MIC, mic, mic);;
	TFMatrix<HT_COMPLEX_DATA_TYPE> matrix2_copy = matrix2->range(HT::MUSIC::UIData::DM3, HT::MUSIC::UIData::MIC, mic, mic);
	matrix1_copy.abs();
	matrix2_copy.abs();
	// 差分
	matrix1_copy.range(HT::MUSIC::UIData::DM3, HT::MUSIC::UIData::MIC, mic, mic)-=matrix2_copy.range(HT::MUSIC::UIData::DM3, HT::MUSIC::UIData::MIC, mic, mic);

	// グラフオブジェクト生成
	int nmic, nsrc, fft, fs;
	(void)nsrc;
	(void)nmic;
	HT::Transform transform;
	transform.BasicData(*data, &nmic, &nsrc, &fft, &fs);
	GNUPlot<std::complex<double>, 3> graph;
	matrix1_copy.createBoostMatrix();

	vector<int> indexes(3);
	indexes[0] = GNUPlot<>::AXIS::X; // FFT
	indexes[1] = mic; // MIC 注意
	indexes[2] = GNUPlot<>::AXIS::Y; // SRC
	graph.setMatrix(matrix1_copy.boostMatrix(), indexes);

	// タイトル
	graph.setTitle("Difference of Amplitude");
	graph.setZUnit(GNUPlot<>::ZUnit::REAL);

	// スタイル
	graph.setStyle(GNUPlot<>::PlotStyle::LINES, GNUPlot<>::Surface::LINE,  GNUPlot<>::ColorMap::UNDEFINE);

	// 軸のラベル
	std::vector<std::string> labels(4);
	labels[0] = "f[Hz]";
	labels[1] = "Mic[No.]";
	labels[2] = "Src[No.]";
	labels[3] = "Amplitude";

	graph.setLabels(labels);

	// 軸の重み
	std::vector<double> weights(4);
	weights[0] = double(fs) / fft;
	weights[1] = 1; // MIC
	weights[2] = 1; // SRC
	weights[3] = 1;
	graph.setWeights(weights);

	// グラフのプロット
	graph.plot(harktoolDir());

	return;
}

void GUIVerifyPanel::dispDiffGraph() {

	QWidget* widget1 = ui->file1->itemData(ui->file1->currentIndex()).value<QWidget*> ();
	QWidget* widget2 = ui->file1->itemData(ui->file2->currentIndex()).value<QWidget*> ();


	GUITabDocument* guidoc1 = dynamic_cast<GUITabDocument*>(widget1);
	HT::UIDocument* doc1 = guidoc1->getDoc();

	GUITabDocument* guidoc2 = dynamic_cast<GUITabDocument*>(widget2);
	HT::UIDocument* doc2 = guidoc2->getDoc();

	HT::UIData *data = doc1->getBaseUIData();
	HT::Transform transform1;
	TFMatrix<HT_COMPLEX_DATA_TYPE> *matrix1 = &transform1.commonTFMatrix(*data);

	HT::UIData *data2 = doc2->getBaseUIData();
	HT::Transform transform2;
	TFMatrix<HT_COMPLEX_DATA_TYPE> *matrix2 = &transform2.commonTFMatrix(*data2);

	dispComparedGraph(matrix1, matrix2, data);

}

void GUIVerifyPanel::targetChanged() {

	try {

		QObject *obj = sender();
		if (obj == ui->file_type) {
			selectCombo(ui->file1, 0);
			selectCombo(ui->file2, 0);

			GUIMainWindow* main = GUIMainWindow::getMainWindow();
			int id = ui->file_type->itemData(ui->file_type->currentIndex()).toInt();
			QTabWidget* tab = main->secondTabWidget((HT::UIApp::GROUPID)id, false);
			removeComboAll(ui->file1);
			removeComboAll(ui->file2);

			ui->file1->addItem("(Undefined)");
			ui->file2->addItem("(Undefined)");
			if (tab) {
				for (int i = 0; i < tab->count(); i++) {
					QString text = tab->tabText(i);
					ui->file1->addItem(text, qVariantFromValue(tab->widget(i)));
					ui->file2->addItem(text, qVariantFromValue(tab->widget(i)));
				}
			}
		}
		if (ui->file1->currentText() != "(Undefined)"
				&& ui->file2->currentText() != "(Undefined)"
				&& ui->file1->currentText() != ui->file2->currentText()) {
			ui->exec_compare->setDisabled(false);
		} else {
			ui->exec_compare->setDisabled(true);
		}
		ui->result_box->hide();
		adjustSize();

	} catch (std::exception& exp) {
		DISP_DIALOG(exp);
		return;
	}
}

GUIVerifyPanel::~GUIVerifyPanel() {
	delete ui;
}

void GUIVerifyPanel::changeEvent(QEvent *e) {
	QDialog::changeEvent(e);
	switch (e->type()) {
	case QEvent::LanguageChange:
		ui->retranslateUi(this);
		break;
	default:
		break;
	}
}
