#ifndef MATRIX_H
#define MATRIX_H

#include <cstdarg>
#include <iostream>
#include <string>
#include <typeinfo>
#include <complex>
#include "UIException.h"
#include "common.h"

#include <boost/multi_array.hpp>
#include <boost/array.hpp>


#if 0 //(sn) one line added
using namespace std;
#endif


//! 3次元ベクトルの格納クラス
/*!
 * 本来はＮ次元ベクトルの格納クラスを目標として作ったが、現状、３次元までしか対応していないが、
 * ほとんどのインターフェース関数は、Ｎ次元に対応出来るように設計されている。
 */
template <typename T>
class TFMatrix {
private:
	std::vector<int> dimension;	//! マトリックスの各次元の深さを表す数字の配列
	//! matrixに対する演算の範囲を決める
	/*!
	 * range()関数によってoperator_rangeはセットされる
	 * range関数与えるデータは可変引数が普通である。
	 */
	std::vector<std::pair<int,int> > operator_range;


	//! operatorの演算の最初に operator_rangeの初期化処理を行う。
	//! そうしないと、演算子の両辺のマトリックスサイズに関してシビアになる
	void init_operator_range(TFMatrix<T> *lhs, TFMatrix<T> *rhs=NULL) {

		// 左辺のMatrixサイズが0の時は、特別に右辺を参考にしてリサイズを行う
		if (rhs && lhs->size().size() == 0) {
			if (rhs->operator_range.size() == 0) {
					lhs->resize(rhs->size());
			} else {
				lhs->resize_by_range(rhs->operator_range);
			}
		}
		// operator_rangeが0の時は、必ず最大範囲に初期化される
		if (lhs->operator_range.size() == 0) {
			lhs->operator_range = lhs->all_range();
		}
		if (rhs && rhs->operator_range.size() == 0) {
			rhs->operator_range = rhs->all_range();
		}
	}

	//! operatorの演算の最後に、operator_range後処理を行う。そうしないと、次の演算時に影響が出てしまう
	void finish_operator_range(TFMatrix<T> *lhs, TFMatrix<T> *rhs=NULL) {
		lhs->operator_range.clear();
		if (rhs) {
			rhs->operator_range.clear();
		}
	}

public:
	TFMatrix() {
		boost_matrix = NULL;
		dimension.clear();
	}

	TFMatrix(int argc, ...) {
		boost_matrix = NULL;
		va_list ap;
		va_start(ap, argc);
		vresize(argc, ap);
		va_end(ap);
	}

	~TFMatrix() {
	}

	// boostのマトリックスに変換する
	boost::multi_array<T, 3>* createBoostMatrix() {
		std::vector<std::pair<int,int> >range_list;
		range_list = all_range();

		delete boost_matrix;
		int range1 = range_list[0].second - range_list[0].first + 1;
		int range2 = range_list[1].second - range_list[1].first + 1;
		int range3 = range_list[2].second- range_list[2].first + 1;
		boost::multi_array<T, 3> *boost_temp = new boost::multi_array<T, 3> (boost::extents[range1][range2][range3]);

		for (int i = range_list[0].first; i <= range_list[0].second; i++) {
			for (int j = range_list[1].first; j <= range_list[1].second; j++) {
				for (int k = range_list[2].first; k <= range_list[2].second; k++) {
					(*boost_temp)[i][j][k] = matrix[i][j][k];
				}
			}
		}
		return this->boost_matrix = boost_temp;
	}

	// boostのマトリックスの取得
	boost::multi_array<T, 3>* boostMatrix() {
		return boost_matrix;
	}

	//! マトリックスの本体データ。現状３次元のvectorで作成されている
	std::vector<std::vector<std::vector<T> > > matrix;
	boost::multi_array<T, 3>* boost_matrix;

	//! \name デバック関数
	//\{
	//! 演算子のオーバーロードで計算する際に、左右両辺の次元が異なるかをチェックする
	void check_dimension(TFMatrix<T> *lhs, TFMatrix<T> *rhs);

	//! 各次元とその深さを表示する
	void print_size() {
		for(unsigned int i = 0; i < dimension.size(); i++) {
			std::cout << "dim=" << i << " depth=" << dimension[i] << std::endl;
		}
		std::cout << std::endl;
	}

	//! 各次元毎に、計算範囲(range)を表示する
	void print_range(std::vector<std::pair<int,int> >range_list) {
		for(unsigned int i = 0; i < range_list.size(); i++) {
			std::cout << "dim=" << i << " range=" << range_list[i].first << "-" << range_list[i].second << std::endl;
		}
	}

	//! 全てのマトリックス要素を表示する
	void print() {
		std::vector<std::pair<int,int> >range_list;
		range_list = all_range();
		print(range_list);
	}

	//! rangeで指定された範囲のマトリックス要素を表示する
	void print(std::vector<std::pair<int,int> >range_list) {
#if 0 //(sn)
cout << "(sn) "
		<< range_list[0].first << " " << range_list[0].second << " " 
		<< range_list[1].first << " " << range_list[1].second << " "
		<< range_list[2].first << " " << range_list[2].second << " " 
		<< endl;
#endif

		for(int i = range_list[0].first; i <= range_list[0].second; i++) {
			for(int j = range_list[1].first; j <= range_list[1].second; j++) {
				for(int k = range_list[2].first; k <= range_list[2].second; k++) {
					std::cout << i << "," << j << "," << k << "=" << matrix[i][j][k] << std::endl;
				}
			}
		}
	}

	//! NaN と Inifinityが含まれるかをチェックする
	int checkInfinite(bool print=false) {
		std::vector<std::pair<int,int> >range_list;
		range_list = all_range();
		int count =0;
		for(int i = range_list[0].first; i <= range_list[0].second; i++) {
			for(int j = range_list[1].first; j <= range_list[1].second; j++) {
				for(int k = range_list[2].first; k <= range_list[2].second; k++) {
					// absを挟んでいるので、実数にも虚数にも両対応
					if (!isfinite(std::abs(matrix[i][j][k]))) {
						if (print) std::cout << i << "," << j << "," << k << "=" << matrix[i][j][k] << std::endl;
						count++;
					}
				}
			}
		}
		if (print) std::cout << "Infinite Total Count " << count << std::endl;
		return count;
	}
	//\}

	
public:
	//! \name マトリックスのサイズを変える関数一覧
	/*!
	 * マトリックスのサイズが強制的にリサイズされるだけで内容に関しては保障されないことに注意する
	 * 事実上、クリア状態からのリサイズに使用する。
	 */
	//\{
	//! 可変引数で変更するサイズを与える
	void resize(int argc, ...) {
		std::vector<int> size;
		// 引数処理
		va_list ap;
		va_start(ap, argc);
		vresize(argc, ap);
		va_end(ap);
	}

	//! マトリックスのサイズ自体を０にする
	void clear() {
		matrix.clear();
		operator_range.clear();
		dimension.clear();
	}

private:
	//! resize のva_listを使う内部関数
	void vresize(int argc, va_list ap) {
		std::vector<int> size_list;
		// 引数処理
		size_list = varg_vector(argc, ap);
		resize(size_list);
	}

	//! resize の可変引数を使用しない内部関数
	/*!
	 * rangeで指定された範囲と同じ大きさになるだけで、
	 * その範囲の要素の内容は保持されない。
	 */
	void resize_by_range(std::vector<std::pair<int, int> > range) {
		std::vector<int> list;

		list.resize(range.size());
		for (unsigned int i=0; i<range.size(); i++) {
			list[i] = range[i].second - range[i].first + 1;
		}
		resize(list);
	}

	//! 最終的にこの関数が呼ばれてリサイズが行われる
	void resize(std::vector<int> size) {
		matrix.resize(size[0]);
		for(int i = 0; i < size[0]; i++) {
			matrix[i].resize(size[1]);
			for(int j = 0; j < size[1]; j++) {
				matrix[i][j].resize(size[2]);
			}
		}
		dimension = size; // 注意
	}

	//! operator_rangeをクリアする関数(内部的に使う関数)
	void clear_range() {
		operator_range.clear();
	}
	//\}

public:
	//! 単に各次元の深さが入った配列を返す
	std::vector<int> size() {
		return dimension;
	}

public:
	//! \name 切り出し関数
	//\{
	//!マトリックスの指定された領域を切りだして取得。
	/*! operator_range関数で領域を指定するのではない事に注意
	 * 右辺値にはなれない事に注意する、領域をコピーするには左辺値のマトリックスの operator_rangeを空にして代入演算を行う
	 */
	void trim(int argc, ...) {
		va_list ap;
		va_start(ap, argc);
		trim(varg_range(argc, ap));
		va_end(ap);
		return;
	}

	//!マトリックスの指定された次元の指定範囲を切り捨てる
	/*!ある次元の全範囲を削除すると、全次元を削除した事になってしまうので削除できない
	 * rangeが全範囲指定の場合、その次元に対して、削除を行わない事を意味する
	 */
	void cut_trim(int argc, ...) {
		va_list ap;
		va_start(ap, argc);
		cut_trim(varg_range(argc, ap));
		va_end(ap);
		return;
	}

private:
	void trim(std::vector<std::pair<int,int> >range_list) {
		// リサイズ元をテンポラリーにコピー
		TFMatrix<T> tmp_matrix;
		tmp_matrix = (*this);

		// リサイズ元をリサイズ
		this->resize_by_range(range_list);

		// コピーからリサイズ元に戻す
		for(int i = range_list[0].first; i <= range_list[0].second; i++) {
			for(int j = range_list[1].first; j <= range_list[1].second; j++) {
				for(int k = range_list[2].first; k <= range_list[2].second; k++) {
					matrix[i-range_list[0].first][j-range_list[1].first][k-range_list[2].first] = tmp_matrix.matrix[i][j][k];
				}
			}
		}
	}

	void cut_trim(std::vector<std::pair<int,int> >range_list) {
		for(unsigned int dim=0; dim < range_list.size(); dim++) {
			int from = range_list[dim].first;
			int to   = range_list[dim].second;
			switch(dim) {
			case(0):
			{
				if (from == 0 &&   to + 1 >= dimension[dim]) {
					continue; // 各次元の全てが指定された場合は実行を行わない
				}
				typename std::vector<std::vector<std::vector<T> > >::iterator it = matrix.begin();
				it+=from;
				for(int cnt=0; cnt < to - from + 1; cnt++) {
					it = matrix.erase(it);
				}
				dimension[0]-=(to-from+1); // サイズの更新
				break;
			}
			case(1):
			{
				if (from == 0 &&   to + 1 >= dimension[dim]) {
					continue; // 各次元の全てが指定された場合は実行を行わない
				}
				for(int i = 0; i < dimension[0]; i++) {
					typename std::vector<std::vector<T> >::iterator it = matrix[i].begin();
					it+=from;
					for(int cnt=0; cnt < to - from + 1; cnt++) {
						it = matrix[i].erase(it);
					}
				}
				dimension[1]-=(to-from+1); // サイズの更新
				break;
			}
			case(2):
			{
				if (from == 0 &&   to + 1 >= dimension[dim]) {
					continue; // 各次元の全てが指定された場合は実行を行わない
				}
				for(int i = 0; i < dimension[0]; i++) {
					for(int j = 0; j < dimension[1]; j++) {
						typename  std::vector<T>::iterator it = matrix[i][j].begin();
						it+=from;
						for(int cnt=0; cnt < to - from + 1; cnt++) {
							it = matrix[i][j].erase(it);
						}
					}
				}
				dimension[2]-=(to-from+1); // サイズの更新
				break;
			}
			}
		}
	}			
	//\}
			
public:
	//! operatorの演算対象の範囲をセットする。
	TFMatrix<T> &range(int argc, ...) {
		va_list ap;
		va_start(ap, argc);
		operator_range = varg_range(argc, ap);
		va_end(ap);
		return (*this);
	}

private:
	//! 引数リストからrangeを生成する
	std::vector<std::pair<int,int> > varg_range(int argc, va_list ap) {
		// 戻り値
		std::vector<std::pair<int,int> > result;

		// 全範囲を格納
		result = all_range();

		// 引数不一致
		if (argc%3) { // A,from,to A次元目がfromからtoまで
			return result;
		}

		int dim;
		for(int i = 0; i < argc; i+=3) {
			dim = va_arg(ap, int);
			result[dim].first  = va_arg(ap, int);
			result[dim].second = va_arg(ap, int);
		}
#if 0 //(sn)表示が多い
cout << "(sn) Itta-140 Matrix.h varg_range argc=" << argc << " dim=" << dim << " " << 
		result[0].first << " " << result[0].second << " " << 
		result[1].first << " " << result[1].second << " " <<
		result[2].first << " " << result[2].second << endl;
#endif
		return result;
	}

	//! 引数リストから、ベクターを生成する
	std::vector<int> arg_vector(int argc, ...) {
		std::vector<int> result;

		va_list ap;
		va_start(ap, argc);
		result = varg_vector(argc, ap);
		va_end(ap);

		return ap;
	}

	//! 引数リストから、ベクターを生成する
	std::vector<int> varg_vector(int argc, va_list ap) {
		// 戻り値
		std::vector<int> result;
		result.resize(argc);

		for(int i = 0; i < argc; i++) {
			result[i] = va_arg(ap, int);
		}
		return result;
	}

	TFMatrix<T> &range(std::vector<std::pair<int,int> >range) {
		operator_range = range;
		return (*this);
	}

	//! operator_range自体の取得
	std::vector<std::pair<int,int> > get_range() {
		return operator_range;
	}

	//! 全範囲のrange値を取得
	std::vector<std::pair<int,int> > all_range() {
		std::vector<std::pair<int,int> > result;
			result.resize(dimension.size());

			// 初期値として全領域を与える
			for (unsigned int i = 0; i < dimension.size(); i++) {
				result[i].first  = 0;
				result[i].second = dimension[i] -1;
			}
			return result;
	}

	//! 引数リストからrangeを生成する
	/*! 例 arg_range(6,1,0,3,2,1,1); // 引数6個 1次元目は0〜3 2次元目は1〜1 それ以外の次元は全範囲
	 */
	std::vector<std::pair<int,int> > arg_range(int argc, ...) {
		std::vector<std::pair<int,int> > result;

		va_list ap;
		va_start(ap, argc);
		result = varg_range(argc, ap);
		va_end(ap);

		return result;
	}


public:
	//! "=マトリックス値"演算子のオーバーロード　単項演算子である事に注意する
	TFMatrix<T> &operator=(TFMatrix<T> &rhs) {
		TFMatrix<T> &lhs = (*this);
		init_operator_range(&lhs, &rhs);
		//IFDEBUG(check_dimension(&lhs, &rhs));

		// 暫定処置
		int i_start =lhs.operator_range[0].first;
		int j_start =lhs.operator_range[1].first;
		int k_start =lhs.operator_range[2].first;
		int rhs_i_start = rhs.operator_range[0].first;
		int rhs_j_start = rhs.operator_range[1].first;
		int rhs_k_start = rhs.operator_range[2].first;

		for(int i = operator_range[0].first; i <= operator_range[0].second; i++) {
			for(int j = operator_range[1].first; j <= operator_range[1].second; j++) {
				for(int k = operator_range[2].first; k <= operator_range[2].second; k++) {
					lhs.matrix[i][j][k] = rhs.matrix[i-i_start + rhs_i_start][j-j_start + rhs_j_start][k-k_start + rhs_k_start];
				}
			}
		}
		finish_operator_range(&lhs, &rhs);
		return lhs;
	}

	//! "=スカラー値"演算子のオーバーロード　単項演算子である事に注意する
	template<typename T2> TFMatrix<T> &operator=(T2 rhs) {
		TFMatrix<T> &lhs = (*this);
		init_operator_range(&lhs);

		for(int i = operator_range[0].first; i <= operator_range[0].second; i++) {
			for(int j = operator_range[1].first; j <= operator_range[1].second; j++) {
				for(int k = operator_range[2].first; k <= operator_range[2].second; k++) {
					lhs.matrix[i][j][k] = rhs;
				}
			}
		}
		finish_operator_range(&lhs);

		return lhs;
	}

	//! "-=マトリックス値"演算子のオーバーロード　単項演算子である事に注意する
	TFMatrix<T> &operator-=(TFMatrix<T> &rhs) {
		TFMatrix<T> &lhs = (*this);
		init_operator_range(&lhs, &rhs);
		//IFDEBUG(check_dimension(&lhs, &rhs));

		// 暫定処置、出来れば、vectorを使わないn次元の形式にしたい
		int i_start =lhs.operator_range[0].first;
		int j_start =lhs.operator_range[1].first;
		int k_start =lhs.operator_range[2].first;
		int rhs_i_start = rhs.operator_range[0].first;
		int rhs_j_start = rhs.operator_range[1].first;
		int rhs_k_start = rhs.operator_range[2].first;

		for(int i = operator_range[0].first; i <= operator_range[0].second; i++) {
			for(int j = operator_range[1].first; j <= operator_range[1].second; j++) {
				for(int k = operator_range[2].first; k <= operator_range[2].second; k++) {
					lhs.matrix[i][j][k] -= rhs.matrix[i-i_start + rhs_i_start][j-j_start + rhs_j_start][k-k_start + rhs_k_start];
				}
			}
		}
		finish_operator_range(&lhs, &rhs);

		return lhs;
	}

	//! "-=スカラー値"演算子のオーバーロード　単項演算子である事に注意する
	template<typename T2> TFMatrix<T> &operator-=(T2 rhs) {
		TFMatrix<T> &lhs = (*this);
		init_operator_range(&lhs);

		for(int i = operator_range[0].first; i <= operator_range[0].second; i++) {
			for(int j = operator_range[1].first; j <= operator_range[1].second; j++) {
				for(int k = operator_range[2].first; k <= operator_range[2].second; k++) {
					lhs.matrix[i][j][k] -= rhs;
				}
			}
		}
		finish_operator_range(&lhs);

		return lhs;
	}

/*
 *  =の付いてない演算子を作ってみたが、問題が多い
// OPERATOR/digit
	template<typename T2> TFMatrix<T> &operator/(T2 rhs) {

		TFMatrix<T> &lhs = (*this);
		init_operator_range(&lhs);

		TFMatrix<T> temp;
		std::vector<std::pair<int,int> > backup_range = lhs.get_range();
		temp = lhs.range(lhs.all_range());
		temp.range(backup_range).operator/=(rhs);

		finish_operator_range(&lhs);
// HIGH COST
		static TFMatrix<T> static_temp;
		static_temp.clear();
		static_temp = temp;
		return static_temp;
	}
*/

	//! "/=マトリックス値"演算子のオーバーロード　単項演算子である事に注意する
	TFMatrix<T> &operator/=(TFMatrix<T> &rhs) {
		TFMatrix<T> &lhs = (*this);
		init_operator_range(&lhs, &rhs);

#if 0 // (sn) Six line added for debug
cout << "(sn) &operator/=(TFMatrix<T> "
		<< operator_range[0].first << " " << operator_range[0].second << " " 
		<< operator_range[1].first << " " << operator_range[1].second << " "
		<< operator_range[2].first << " " << operator_range[2].second << " " 
		<< endl;
#endif

		for(int i = operator_range[0].first; i <= operator_range[0].second; i++) {
			for(int j = operator_range[1].first; j <= operator_range[1].second; j++) {
				for(int k = operator_range[2].first; k <= operator_range[2].second; k++) {
					lhs.matrix[i][j][k] /= rhs.matrix[i][j][k];
				}
			}
		}
		finish_operator_range(&lhs, &rhs);
		return lhs;
	}

	//! "/=スカラー値"演算子のオーバーロード　単項演算子である事に注意する
	template<typename T2> TFMatrix<T> &operator/=(T2 rhs) {
		TFMatrix<T> &lhs = (*this);
		init_operator_range(&lhs);

#if 0 // (sn) Six line added for debug
cout << "(sn) Itta-45 "
		<< operator_range[0].first << " " << operator_range[0].second << " " 
		<< operator_range[1].first << " " << operator_range[1].second << " "
		<< operator_range[2].first << " " << operator_range[2].second << " " 
		<< rhs
		<< endl;
#endif

		for(int i = operator_range[0].first; i <= operator_range[0].second; i++) {
			for(int j = operator_range[1].first; j <= operator_range[1].second; j++) {
				for(int k = operator_range[2].first; k <= operator_range[2].second; k++) {
					lhs.matrix[i][j][k] /= rhs;
				}
			}
		}
		finish_operator_range(&lhs);


		return lhs;
	}

	//! "*=マトリックス値"演算子のオーバーロード　単項演算子である事に注意する
	TFMatrix<T> &operator*=(TFMatrix<T> &rhs) {
		TFMatrix<T> &lhs = (*this);
		init_operator_range(&lhs, &rhs);

		for(int i = operator_range[0].first; i <= operator_range[0].second; i++) {
			for(int j = operator_range[1].first; j <= operator_range[1].second; j++) {
				for(int k = operator_range[2].first; k <= operator_range[2].second; k++) {
					lhs.matrix[i][j][k] *= rhs.matrix[i][j][k];
				}
			}
		}
		finish_operator_range(&lhs, &rhs);
		return lhs;
	}

	//! "*=スカラー値"演算子のオーバーロード　単項演算子である事に注意する
	template<typename T2> TFMatrix<T> &operator*=(T2 rhs) {
		TFMatrix<T> &lhs = (*this);
		init_operator_range(&lhs);

		for(int i = operator_range[0].first; i <= operator_range[0].second; i++) {
			for(int j = operator_range[1].first; j <= operator_range[1].second; j++) {
				for(int k = operator_range[2].first; k <= operator_range[2].second; k++) {
					lhs.matrix[i][j][k] *= rhs;
				}
			}
		}
		finish_operator_range(&lhs);

		return lhs;
	}


	//! 位相 Complex専用関数である事に注意する
	TFMatrix<T> &arg() {
		if(get_range().size() == 0) {
			operator_range = all_range();
		}
		std::vector<std::pair<int,int> > &range_list = operator_range;
		for(int i = range_list[0].first; i <= range_list[0].second; i++) {
			for(int j = range_list[1].first; j <= range_list[1].second; j++) {
				for(int k = range_list[2].first; k <= range_list[2].second; k++) {
					// argはcomplexの専用関数
//					if(std::abs(matrix[i][j][k].real()) < 0.0001 || std::abs(matrix[i][j][k].imag()) < 0.001) {
//						matrix[i][j][k] = 0;
//					} else {
						matrix[i][j][k] = std::arg(matrix[i][j][k]);
//					}
				}
			}
		}
		clear_range();
		return *this;
	}

	//! dB(Log10*abs*10) この関数はComplex専用の関数である事に注意
	TFMatrix<T> &db() {
		if(get_range().size() == 0) {
			operator_range = all_range();
		}
		std::vector<std::pair<int,int> > &range_list = operator_range;
		for(int i = range_list[0].first; i <= range_list[0].second; i++) {
			for(int j = range_list[1].first; j <= range_list[1].second; j++) {
				for(int k = range_list[2].first; k <= range_list[2].second; k++) {
					T &temp = matrix[i][j][k];
					// abs,real()などcomplexの関数が呼び出されている
					temp = std::abs(temp);
					temp = std::log10(temp.real()) * 10;
					matrix[i][j][k] = temp;
				}
			}
		}
		clear_range();
		return *this;
	}

	//! 合計
	T sum() {
		std::vector<std::pair<int,int> >temp;
		if(get_range().size() == 0) {
			temp = all_range();
		} else {
			temp = operator_range;
		}
		clear_range();
		std::vector<std::pair<int,int> > &range_list = temp;

		T total = 0;
		for(int i = range_list[0].first; i <= range_list[0].second; i++) {
			for(int j = range_list[1].first; j <= range_list[1].second; j++) {
				for(int k = range_list[2].first; k <= range_list[2].second; k++) {
					total+=(matrix[i][j][k]);
				}
			}
		}

#if 0 //(sn)たくさん表示される
cout << "(sn) T sum() "
		<< range_list[0].first << " " << range_list[0].second << " " 
		<< range_list[1].first << " " << range_list[1].second << " " 
		<< range_list[2].first << " " << range_list[2].second << " " 
		<< total
		<< endl;
#endif
		return total;
	}

	//! 相加平均
	T mean() {
		std::vector<std::pair<int,int> >temp;
		if(get_range().size() == 0) {
			temp = all_range();
		} else {
			temp = operator_range;
		}
		clear_range();
		std::vector<std::pair<int,int> > &range_list = temp;

		T total = 0;
		int sum_count =0;

		for(int i = range_list[0].first; i <= range_list[0].second; i++) {
			for(int j = range_list[1].first; j <= range_list[1].second; j++) {
				for(int k = range_list[2].first; k <= range_list[2].second; k++) {
					total+=(matrix[i][j][k]);
					sum_count++;
				}
			}
		}

#if 0 //(sn)たくさん表示される
cout << "(sn) T mean() "
		<< range_list[0].first << " " << range_list[0].second << " " 
		<< range_list[1].first << " " << range_list[1].second << " " 
		<< range_list[2].first << " " << range_list[2].second << " " 
		<< total << " " << sum_count
		<< endl;
#endif
		return total/(T)sum_count;
	}

	//! 絶対値 複素数でもＯＫ
	TFMatrix<T> &abs() {
		if(get_range().size() == 0) {
			operator_range = all_range();
		}

		std::vector<std::pair<int,int> > &range_list = operator_range;

#if 0
cout << "(sn) "
		<< range_list[0].first << " " << range_list[0].second << " " 
		<< range_list[1].first << " " << range_list[1].second << " "
		<< range_list[2].first << " " << range_list[2].second << " " 
		<< endl;
#endif

		for(int i = range_list[0].first; i <= range_list[0].second; i++) {
			for(int j = range_list[1].first; j <= range_list[1].second; j++) {
				for(int k = range_list[2].first; k <= range_list[2].second; k++) {
					matrix[i][j][k] = std::abs(matrix[i][j][k]);
				}
			}
		}
		return *this;
	}

	//! 最大値 
	/*!
	 * operator_range範囲の最大値を返す
	 * offsetには、配列の添え字が戻るが、２次元以上になった時のoffsetの数え方は、MATLABと異なるので注意する
	 * MAXとなる同値の数字があった場合は、最も小さいoffsetが戻る
	 * 戻り値には、対象となる要素の値が戻る
	 */
	T complex_max(int *offset=NULL) {
		if(get_range().size() == 0) {
			operator_range = all_range();
		}
		T max_temp = matrix[0][0][0]; // point
		int max_offset = 0;
		int count=0;
		std::vector<std::pair<int,int> > &range_list = operator_range;
		for(int i = range_list[0].first; i <= range_list[0].second; i++) {
			for(int j = range_list[1].first; j <= range_list[1].second; j++) {
				for(int k = range_list[2].first; k <= range_list[2].second; k++) {
					if (matrix[i][j][k].real() > max_temp.real()) {
						max_temp = matrix[i][j][k];
						max_offset=count;
					}
					count++;
				}
			}
		}
		if(offset) {
			*offset = max_offset;
		}
		return max_temp;
	}
	
	
	
	//! 複素共役 (単項演算子)
	TFMatrix<T> &conj() {
		if(get_range().size() == 0) {
			operator_range = all_range();
		}
		std::vector<std::pair<int,int> > &range_list = operator_range;
		for(int i = range_list[0].first; i <= range_list[0].second; i++) {
			for(int j = range_list[1].first; j <= range_list[1].second; j++) {
				for(int k = range_list[2].first; k <= range_list[2].second; k++) {
					matrix[i][j][k] = std::conj(matrix[i][j][k]);
				}
			}
		}
		clear_range();
		return *this;
	}

	//! べき乗(単項演算子)
	TFMatrix<T> &pow(int index) {
		if(get_range().size() == 0) {
			operator_range = all_range();
		}
		std::vector<std::pair<int,int> > &range_list = operator_range;
		for(int i = range_list[0].first; i <= range_list[0].second; i++) {
			for(int j = range_list[1].first; j <= range_list[1].second; j++) {
				for(int k = range_list[2].first; k <= range_list[2].second; k++) {
					matrix[i][j][k] = std::pow(matrix[i][j][k], index);
				}
			}
		}
		clear_range();
		return *this;
	}

	//! 逆正接(単項演算子)
	TFMatrix<T> &atan2(TFMatrix<T> &rhs) {
		TFMatrix<T> &lhs = (*this);
		init_operator_range(&lhs, &rhs);
		IFDEBUG(check_dimension(&lhs, &rhs));

		// 暫定処置
		int i_start =lhs.operator_range[0].first;
		int j_start =lhs.operator_range[1].first;
		int k_start =lhs.operator_range[2].first;
		int rhs_i_start = rhs.operator_range[0].first;
		int rhs_j_start = rhs.operator_range[1].first;
		int rhs_k_start = rhs.operator_range[2].first;

		for(int i = operator_range[0].first; i <= operator_range[0].second; i++) {
			for(int j = operator_range[1].first; j <= operator_range[1].second; j++) {
				for(int k = operator_range[2].first; k <= operator_range[2].second; k++) {
					lhs.matrix[i][j][k] = atan2(lhs.matrix[i][j][k], rhs.matrix[i-i_start + rhs_i_start][j-j_start + rhs_j_start][k-k_start + rhs_k_start]);
				}
			}
		}
		finish_operator_range(&lhs, &rhs);
		return *this;
	}

/*
// range_list を文字列で与える 　試作
// 例: range_parser(:,1:2,2,) 等
	std::vector<std::pair<int,int> > range_parser(string range_str) {

		// range_listは最初にサイズの初期化する
		// 初期値として全領域を与える
		std::vector<std::pair<int,int> > range_list;
		range_list = all_range();

		int last_offset = 0;
		int dimension_depth =0;
		int find_offset;
		for(int dim_loop=0; ; dim_loop++) {
			find_offset  = range_str.find(",", last_offset);

			if(find_offset < 0) {
				find_offset = range_str.length();
			}
			string split = range_str.assignr(last_offset, find_offset - last_offset);
			last_offset  = find_offset + 1;

			{
				string range_str = split;
				int find_offset  = range_str.find(":", 0);
				if (find_offset < 0) {
					if (range_str == "") {
						range_list[dim_loop].first =0;
						range_list[dim_loop].second=dimension[dimension_depth] - 1;
					} else {
						int same = atoi(range_str.c_str());
						// ERROR SAFE
						if (same < 0) {
							same = 0;
						}
						if (same >= dimension[dimension_depth]) {
							same = dimension[dimension_depth] - 1;
						}
						range_list[dim_loop].first =same;
						range_list[dim_loop].second=same;
					}
				} else {
					string lhs = range_str.assignr(0, find_offset);
					string rhs = range_str.assignr(find_offset + 1);
					int int_lhs, int_rhs;
					// LHS
					if (lhs == "") {
						int_lhs = 0 ;
					} else {
						int_lhs = atoi(lhs.c_str());
					}
					// RHS
					if (rhs == "") {
						int_rhs = dimension[dimension_depth] -1;
					} else {
						int_rhs = atoi(rhs.c_str());
					}
					// LHS ERROR SAFE
					if (int_lhs < 0) {
						int_lhs = 0;
					}
					if (int_lhs >= dimension[dimension_depth]) {
						int_lhs = dimension[dimension_depth] - 1;
					}
					// RHS ERROR SAFE
					if (int_rhs < 0) {
						int_rhs = 0;
					}
					if (int_rhs >= dimension[dimension_depth]) {
						int_rhs = dimension[dimension_depth] - 1;
					}
					range_list[dim_loop].first = int_lhs;
					range_list[dim_loop].second= int_rhs;
				}
			}
			if (find_offset >= range_str.length()) {
				break;
			}
			dimension_depth++;
		}
		return range_list;
	}
*/
};

#endif
// templateを使ったクラスは、分割コンパイル出来ない為
#include "Matrix.cc"
