#ifndef UIDATDATA_H
#define UIDATDATA_H

#include <iostream>
#include <string>
#include <cmath>
#include <vector>
#include <sstream>
#include <boost/shared_ptr.hpp>
#include <boost/regex.hpp>
#include "UIApp.h"
#include "UIData.h"
#include "UIDatDocument.h"
#include "UIGraph.h"
#include "Matrix.h"
#include "common.h"
#include "libharkio/libharkio.h"


namespace HT {
namespace MUSIC {
class UIDocument;

typedef float HT_FLT_DATA_TYPE;	//! インパルス応答ファイルのデータ型
typedef float HT_MUSIC_DATA_TYPE;	//! MUSICのTFファイルの出力データ型

//! SRCファイル各行の情報を保存するクラス
template <typename T>
class UISrc {
public:
	typedef boost::shared_ptr<UISrc<T> > shared_ptr;
	T x;
	T y;
	T z;
	//! 新規にTFファイルを作成する時だけに使用されるデータ
	//\{
	std::string implement_file;	//!srcinfテキストに埋め込まれたfltファイル名
	int cut_start;				//! srcinfファイルのfltファイル名の後ろの数
	T search_direction;			//! x,y平面の角度の計算結果が保持される。 古いMUSICのファイルフォーマット
	//\}
	//! デカルト座標から、search_directoinを計算する関数
	void setCalcSearchDirection() {
		search_direction = atan2(y, x) / M_PI * 180;
	}
	//! 直接、search_directionを指定してセットする関数
	void setSearchDirection(float val) {
		search_direction = val;
	}
	//! search_directionの値を取得する関数
	T getSearchDirection() {
		return search_direction;
	}
	//!デバックプリント関数
	void print() {
		std::cout << "x=" << x << " y=" << y << "  z=" << z << " file=" << implement_file << " cut_start=" << cut_start << " search_directoin=" << search_direction << std::endl;
	}
};

// メンバ関数用パラメータ
class CreateTFParam {
public:
	std::string mic_file;
	std::string src_file;
	std::string mic_string;
	std::string src_string;

	int NFFT;
	int NFS;
	int DIRECT_LENGTH; // 昔はMUSIC_DIR
	int IMPULSE_RESPONSE_LEN;
};

class moveToHARKIOParam {
public:
	std::vector<HT::UISrc<HT_DATA_TYPE> >* src_list;
	std::vector<HT::UIMic<HT_DATA_TYPE> >* mic_list;
	TFMatrix<HT_COMPLEX_DATA_TYPE>* tf_matrix;
	int NMic;
	int NFFT;
	int NIR; // インパルスレスポンス長
	int NFS;
};

//! MUSICのＴＦファイルのデータコンテナクラス
class UIData : public HT::UIData {
//Q_OBJECT
	bool tsp_canceled;
private:
	// 旧形式データ
	boost::shared_ptr<TFMatrix<HT_COMPLEX_DATA_TYPE> > matrix; //! 伝達関数マトリックス
	boost::shared_ptr<std::vector<HT::UIMic<HT_DATA_TYPE> > > mic; //! マイク座標配列
	boost::shared_ptr<std::vector<HT::UISrc<HT_DATA_TYPE> > > src; //! 音源位置座標配列
	HT_DATA_TYPE search_range; //! Srcの大きさの平均値
	HT_DATA_TYPE search_range_len; // search_rangeの次元サイズを示すが、スカラーデータなので１にしかならない？
	int NFreq;	//! 周波数ライン数
	int NSrc;	//! Src数
	int NMic;	//! マイク数
	int FFT;	//! FFT長
	int FS;		//! サンプリング周波数
	int MusicDir;	//! MUSIC_DIR

	// 新形式 harkio データ
	boost::shared_ptr<harkio::HARK> harkio;
public:
	void setHARKIO(harkio::HARK* harkio) {
		if (this->harkio.get() != harkio) {
			this->harkio.reset(harkio);
		}
	}
	harkio::HARK* getHARKIO() {
		return harkio.get();
	}
	bool isHARKIO() {
		if (harkio.get()) {
			return true;
		} else {
			return false;
		}
	}

	UIDocument* getDocument();
	harkio::HARK* moveToHARKIO(const moveToHARKIOParam& param);

	UIData(UIDocument *document);
	virtual ~UIData();

	//! 伝達関数マトリックスの各次元を表すシンボル
	typedef enum{
		 FRQ = 0
		,MIC = 1
		,SRC = 2
	} MATRIX_SUBSCRIPT;

	//! 伝達関数マトリックス全体の次元数
	typedef enum{DM3 = 3} MATRIX_DIM;

	float getSearchRange() {
		return search_range;
	}
	void setSearchRange(float val) {
		search_range = val;
	}
	//! Src配列から計算によって、search_range をセットする
	void setCalcSearchRange() {
		HT_DATA_TYPE total=0;
		unsigned int count = 0;
		std::vector<HT::UISrc<HT_DATA_TYPE> >* src = getSrc();
		for (count=0; count < src->size(); count++) {
			total += sqrt((*src)[count].x * (*src)[count].x
						+ (*src)[count].y * (*src)[count].y
						+ (*src)[count].z * (*src)[count].z);
		}
		search_range = total/count;
	}

	float getSearchRangeLen() {
		return search_range_len;
	}
	void setSearchRangeLen(int val) {
		search_range_len = val;
	}
	int getNFreq() {
		return NFreq;
	}
	void setNFreq(int value) {
		NFreq = value;
	}

	int getNSrc() {
		return NSrc;
	}
	void setNSrc(int value) {
		NSrc = value;
	}

	int getNMic() {
		return NMic;
	}
	void setNMic(int value) {
		NMic = value;
	}

	int getNFFT() {
		return FFT;
	}
	void setNFFT(int value) {
		FFT = value;
	}

	int getNFS() {
		return FS;
	}
	void setNFS(int value) {
		FS = value;
	}

	void setMusicDir(int value) {
		MusicDir = value;
	}
	int getMusicDir() {
		return MusicDir;
	}

	//! 伝達関数マトリックス Matrix
	template<typename T> void setTFMatrix(T *matrix) {
		if (this->matrix.get() != matrix) {
			this->matrix.reset(matrix);
		}
	}
	TFMatrix<HT_COMPLEX_DATA_TYPE>* getMatrix() {
		return matrix.get();
	}

	//! マイク座標配列 UIMic
	// 旧形式ファイルには、マイク座標は入っていない
	template<typename T> void setMic(T *mic) {
		if (this->mic.get() != mic) {
			this->mic.reset(mic);
		}
	}
	std::vector<HT::UIMic<HT_DATA_TYPE> >* getMic() {
		return mic.get();
	}

	//! 音源座標配列 Src
	template<typename T> void setSrc(T *src) {
		if (this->src.get() != src) {
			this->src.reset(src);
		}
	}
	std::vector<HT::UISrc<HT_DATA_TYPE> >* getSrc() {
		return src.get();
	}

	//! 伝達関数の新規作成関数
	int createTFData(const CreateTFParam& param);
	//! TFファイルとしてデータを保存する関数(旧形式)
	int saveTFFile(std::string fullpath_file);
	//! 逆変換サブ関数
	TFMatrix<HT_COMPLEX_DATA_TYPE>* IFFT(TFMatrix<HT_COMPLEX_DATA_TYPE>& matrix, int FFT, int NMIC, int NSRC);

	void setTSPCanceled(bool value) {
		tsp_canceled = value;
	}
};

} //namespace MUSIC
} //namespace

#endif
