#include "UIDatDocument.h"
#include "UIDatData.h"
#include "UIGraph.h"
#include "UITFDocument.h"
#include "UITFData.h"
#include "Matrix.h"
#include "common.h"
#include "libharkio/libharkio.h"
#include <QObject>
#include <QDateTime>
#include <QString>

#include <cmath>
#include <complex>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/lexical_cast.hpp>

using namespace std;
using namespace boost;
using namespace boost::numeric;
using namespace harkio;

namespace HT {
namespace MUSIC {

UIDocument::UIDocument(std::string name, std::string filename)
{
	setName(name);
	setFileName(filename);
};

UIDocument::UIDocument() {}
UIDocument::~UIDocument() {}

UIData* UIDocument::getUIData(bool exception) {
	if (exception && !data) {
		THROW_CRITICAL("data is null");
	}
	return data.get();
}
HT::UIData* UIDocument::getBaseUIData() {
	return data.get();
}

UIData* UIDocument::newTFData(UIDocument *doc) {
	data.reset(new UIData(doc));
	return data.get();
}

void UIDocument::loadFileByLib(string filename) {
IFDEBUG(DP("# loadFileByLib"));

	HARK *hark = NULL;
	try {
		UIData *data = newTFData(this);		// data生成
		setFileName(filename);

		hark = new HARK;
		hark->readFile(filename);
		data->setHARKIO(hark);

	} catch (harkio::Error &e) {
		delete hark;
		THROW_CRITICAL(e.getMsg().c_str());
	}
}

void UIDocument::loadFileNormally(string filename) {
IFDEBUG(DP("# loadFileNormally"));

	int int_buf;
	int read_size;
	float *float_list = NULL;
	int *int_list = NULL;
	FILE *fp = NULL;

	setFileName(filename);
	UIData *data = newTFData(this); // data生成

	vector<HT::UISrc<HT_DATA_TYPE> > *src = NULL;
	vector<UIMic<HT_DATA_TYPE> > *mic = NULL;
	TFMatrix<HT_COMPLEX_DATA_TYPE> *a_matrix = NULL;

	fp = fopen(filename.c_str(), "rb");
	if (fp == NULL) {
		THROW_WARNING(tr("Can't open file:%1").arg(filename.c_str()));
	}
	try {
		// Nd (SearchDirectionである、新しいNSrc)
		read_size = fread(&int_buf, sizeof(int), 1, fp);
		if (read_size == 1) {
			data->setNSrc(int_buf);
		} else {
			THROW_CRITICAL("Nd read error");
		}

		// Nr
		read_size = fread(&int_buf, sizeof(int), 1, fp);
		if (read_size == 1) {
			data->setSearchRangeLen(int_buf);

		} else {
			THROW_CRITICAL("Nr read error");
		}

		// NMic (マイク配列の長さ)
		read_size = fread(&int_buf, sizeof(int), 1, fp);
		if (read_size == 1) {
			data->setNMic(int_buf);
		} else {
			THROW_CRITICAL("NMic read error");
		}

		// FFT長
		read_size = fread(&int_buf, sizeof(int), 1, fp);
		if (read_size == 1) {
			data->setNFFT(int_buf);
		} else {
			THROW_CRITICAL("FFT read error");
		}
		// search_directionの配列 Nd数の配列長(新しいSrcのx-y平面上の角度?)
		int_list = new int[data->getNSrc()];
		read_size = fread(int_list, sizeof(int), data->getNSrc(), fp);
		if (read_size != data->getNSrc()) {
			THROW_CRITICAL("SearchDirection read error");
		}
		// ここで、UISrc の生成して、SearchDirectionの格納
		src = new vector<HT::UISrc<HT_DATA_TYPE> > ();
		src->resize(data->getNSrc());
		data->setSrc(src); // 保持
		// search_direction だけをコピー
		for (unsigned int i=0; i < (*src).size(); i++) {
			(*src)[i].search_direction = int_list[i];
		}
		delete[] int_list;

		// Search Range 新しいNSrcの２乗平均値
		read_size = fread(&int_buf, sizeof(int), 1, fp);
		if (read_size == 1) {
			data->setSearchRange(int_buf);
		} else {
			THROW_CRITICAL("SearchRange read error");
		}

		// Amatrix 本体読み込み
		int matrix_size = (data->getNSrc()) * (data->getNMic()) * (data->getNFFT() / 2 + 1) * 2; // "* 2"は実部と虚部
		float_list = new float[matrix_size];
		read_size = fread(float_list, sizeof(float), matrix_size, fp);
		if (read_size != matrix_size) {
			delete[] float_list;
			THROW_CRITICAL("AMatrix read error");
		}
		// 終端データの確認
		int end = ftell(fp);
		fseek(fp, 0, SEEK_END);
		if (ftell(fp) != end) {
			delete[] float_list;
			THROW_CRITICAL("Can't reach end of file.");
		}
		// Amatrixコンテナ確保
		a_matrix = new TFMatrix<HT_COMPLEX_DATA_TYPE> (3, data->getNFFT() / 2 + 1, data->getNMic(), data->getNSrc());
		data->setTFMatrix(a_matrix); // マトリックス保持

		int linear_offset = 0;
		for (int nMic = 0; nMic < data->getNMic(); nMic++) {
			for (int nFFT = 0; nFFT < (data->getNFFT() / 2 + 1); nFFT++) {
				for (int nSrc = 0; nSrc < data->getNSrc(); nSrc++) {
					//サイズが 実部、虚部の２倍になる事に注意
					linear_offset = (((data->getNFFT() / 2 + 1) * (data->getNMic() * nSrc) + ((data->getNFFT() / 2 + 1)
							* nMic) + (nFFT)) * 2);
					a_matrix->matrix[nFFT][nMic][nSrc] = HT_COMPLEX_DATA_TYPE(float_list[linear_offset],
							float_list[linear_offset + 1]);
				}
			}
		}
		delete[] float_list;

		// DATファイルからはMIC座標は分からない。マイク数だけは分かる
		// 一応、UIMicだけ生成
		mic = new vector<UIMic<HT_DATA_TYPE> > ();
		mic->resize(data->getNMic());
		data->setMic(mic); // 保持

		// GSSと違って、MUSICの場合は、周波数ラインNFreqと言う情報がない
		data->setNFreq(data->getNFFT() / 2 + 1);
		fclose(fp);

	} catch (...) {
		if (fp != NULL)
			fclose(fp);

		this->data.reset();
		throw;
	}

}

void UIDocument::loadFile(string filename) {

	HARKIO_RET ret = harkio::HARK::checkIdentifier(filename);
	if (ret == harkio::HARKIO_TRUE) {
		loadFileByLib(filename); // ライブラリ読み込み
	} else if (ret == harkio::HARKIO_FALSE) {
		loadFileNormally(filename); // 旧形式の読み込み
	} else {
		THROW_WARNING(harkio::Error::getInstance().getMsg());
	}
}


void UIDocument::mkmusictf(MKMusicParam& param) {

	typedef complex<float> fcomplex;
	typedef ublas::vector<float> fvector;
	typedef ublas::matrix<float> fmatrix;
	typedef ublas::vector<fcomplex> fcvector;
	typedef ublas::matrix<fcomplex> fcmatrix;

	int length = param.NFFT;
	int distance = 100; // [cm]
	int interval = param.DEGREE_INTERVAL; // [degree]
	float speed_of_sound = 343; // [m/s]
	float fixed_elevation = 16.7; // [degree]
	int sampling_rate = param.NFS;
	string mic_filename = param.mic_file;
//	string output_filename;
//	bool exist_mic_filename = false;
//	bool exist_output_filename = false;
	float speed_of_sound_1 = sampling_rate / (100.0 * speed_of_sound); // [sample/cm]
	int flength = length / 2 + 1;
	fmatrix mic_pos;

	// マイクファイル読み込み
	harkio::TextData files;
	harkio::TextBase *file;

	// マイクファイル読み込み
	file = files.readFile(param.mic_file);

	vector<harkio::UIMic<> > *mic_list = &(((harkio::MICARYLocationFile*) file)->mic_lists);
	std::auto_ptr<harkio::TextBase> tempptr(file); // 開放

	mic_pos.resize(mic_list->size(), 3);
	for(uint i=0; i < mic_pos.size1(); i++) {
		mic_pos(i, 0) = (*mic_list)[i].x;
		mic_pos(i, 1) = (*mic_list)[i].y;
		mic_pos(i, 2) = (*mic_list)[i].z;
	}
	int nmic = mic_pos.size1();
	// マイクファイルここまで
	fvector pos_mean(3);
	pos_mean = ublas::zero_vector<float>(3);
	mic_pos *= 100.0f; // convert [m] to [cm]
	for (uint c = 0; c < mic_pos.size1(); c++) {
		pos_mean += row(mic_pos, c);
	}
	pos_mean /= mic_pos.size1();
	for (uint c = 0; c < mic_pos.size1(); c++) {
		row(mic_pos, c) -= pos_mean;
	}
	//cerr << mic_pos << endl;

	vector<int> direction;
	vector<int> range;
	vector<vector<vector<fcvector> > > A;
	range.push_back(distance);
	for (int dir = -180; dir < 180; dir += interval) {
		direction.push_back(dir);
	}

	A.resize(flength);
	for (int k = 0; k < flength; k++) {
		A[k].resize(direction.size());
		for (uint id = 0; id < direction.size(); id++) {
			A[k][id].resize(range.size());
			for (uint ir = 0; ir < range.size(); ir++) {
				A[k][id][ir].resize(nmic);
			}
		}
	}
	vector<fvector> src_pos;
	for (uint id = 0; id < direction.size(); id++) {
		fvector pos(3);
		float angle = direction[id];
		float elevation = fixed_elevation;
		float r = range[0];
		pos[0] = r * cos(angle * M_PI / 180.0) * cos(elevation * M_PI / 180.0);
		pos[1] = r * sin(angle * M_PI / 180.0) * cos(elevation * M_PI / 180.0);
		pos[2] = r * sin(elevation * M_PI / 180.0);

		src_pos.push_back(pos);

		fvector src_deltas(nmic);
		// float d0 = r;
		for (int c = 0; c < nmic; c++) {
			src_deltas[c] = speed_of_sound_1 * (- norm_2(pos - row(mic_pos, c)));
//			src_deltas[c] = speed_of_sound_1 * (d0 - norm_2(pos - row(mic_pos, c)));
		}

		for (int k = 0; k < flength; k++) {
			for (int c = 0; c < nmic; c++) {
				float theta = 2 * M_PI * src_deltas[c] * k / length;
				A[k][id][0][c] = complex<float> (cos(theta), sin(theta));
			}
		}
	}
	// 以下、HARKIOにコピー
	// HGTFデータ
	HARK* hark = new HARK();
	Data *hgtf = new HGTF();
	hark->addData(hgtf);

	//FMTチャンク
	//FMT(FMT::VNAME version, FMT::FORMAT_ID format_id, int nMic, int nFFT, int nIR, int nFS)
	FMT *fmt = new FMT(FMT::V1_0, FMT::M2PG, nmic, length, param.IMPULSE_RESPONSE_LEN, sampling_rate);
	hgtf->addChunk(fmt);

	// M2PGチャンク
	// M2PG(M2PG::VNAME version, FMT *fmt, int nd, int nr);
	M2PG *m2pg = new M2PG(M2PG::V1_0, fmt, direction.size(), 1);
	hgtf->addChunk(m2pg);

	// SPECチャンク
	SPEC *spec = new SPEC(SPEC::V1_0);
	hgtf->addChunk(spec);

	// MICデータ
	FMT::MIC_LIST_TYPE *mic = fmt->newMic(); // マイク座標コンテナ生成
	mic_pos.resize(mic_list->size(), 3);
	for(uint i=0; i < mic_pos.size1(); i++) {
		(*mic)[i].x = (*mic_list)[i].x;
		(*mic)[i].y = (*mic_list)[i].y;
		(*mic)[i].z = (*mic_list)[i].z;
	}

	// 音源位置データ src[nd][nr]
	M2PG::SRC_LIST_TYPE *src = m2pg->getSrc();
	for (uint i = 0; i < src_pos.size(); i++) {
		(*src)[i][0].x = src_pos[i][0];
		(*src)[i][0].y = src_pos[i][1];
		(*src)[i][0].z = src_pos[i][2];
	}

	// 伝達関数データ tf[nd][nr][nmic][nfft]
	M2PG::TF_LIST_TYPE *tf = m2pg->getTF();

	for (uint id = 0; id < direction.size(); id++) {
		for (int k = 0; k < flength; k++) {
			for (int c = 0; c < nmic; c++) {
				(*tf)[id][0][c][k] = A[k][id][0][c];
			}
		}
	}

	// SPECデータ
	spec->setDate(QDateTime::currentDateTime().toString("yyyy/MM/dd").toStdString().c_str());
	spec->setCreator(getUserName().toStdString().c_str());

	UIData* data = getUIData(false);
	data->setHARKIO(hark);

	return;
}

} //namespace MUSIC
} //namespace

