#ifndef UIEXCEPTION_H
#define UIEXCEPTION_H

#include <exception>
#include <string>
#include <QString>

namespace HT {

#define THROW_EXCEPTION(message) \
	throw HT::UIException(message, __FILE__, __func__, __LINE__)

#define THROW_WARNING(message) \
	throw HT::UIWarning(message, __FILE__, __func__, __LINE__)

#define THROW_CRITICAL(message) \
	throw HT::UICritical(message, __FILE__, __func__, __LINE__)

//! 例外、エラー、警告関連のクラス
/*!
 * 例外、エラー処理マクロを使用する為に、このクラスを直接に扱うことは少ない
 * 例外をキャッチした後に、ダイアログを表示するメンバ関数が用意されている。
 */
class UIException : public std::exception {
public:
private:
	QString message; //! エラーメッセージ
	const char* fileName; //! 例外マクロが使用されたファイル名
	const char* functionName; //! 例外マクロが使用された関数名
	int line; //! 例外マクロが使用された行数
public:
	UIException() {}
	UIException(QString aMessage);
	UIException(std::string aMessage);
	UIException(const char* aMessage);

	UIException(QString aMessage,
			const char* aFile,
			const char* aFunction,
			const int aLine);
	UIException(std::string aMessage,
			const char* aFile,
			const char* aFunction,
			const int aLine);
	UIException(const char* aMessage,
			const char* aFile,
			const char* aFunction,
			const int aLine);

	virtual ~UIException() throw();
	const char* getFileName() const;
	const char* getFunctionName() const;
	int getLineNumber() const;
	virtual const char* what() const throw();
	QString& getMessage() {return message;}
};

class UIWarning : public UIException {
public:
	UIWarning(QString aMessage,
		const char* aFile,
		const char* aFunction,
		const int aLine) : UIException(aMessage, aFile, aFunction, aLine) {}
	UIWarning(std::string aMessage,
		const char* aFile,
		const char* aFunction,
		const int aLine) : UIException(aMessage, aFile, aFunction, aLine) {}
	UIWarning(const char* aMessage,
		const char* aFile,
		const char* aFunction,
		const int aLine) : UIException(QString(aMessage), aFile, aFunction, aLine) {}
};

class UICritical : public UIException {
public:
	UICritical(QString aMessage,
		const char* aFile,
		const char* aFunction,
		const int aLine) : UIException(aMessage, aFile, aFunction, aLine){}
	UICritical(std::string aMessage,
		const char* aFile,
		const char* aFunction,
		const int aLine) : UIException(aMessage, aFile, aFunction, aLine){}
	UICritical(const char* aMessage,
		const char* aFile,
		const char* aFunction,
		const int aLine) : UIException(QString(aMessage), aFile, aFunction, aLine){}

};


} // namespace HT


#endif
