#ifndef UIGRAPH_H
#define UIGRAPH_H
#include <boost/multi_array.hpp>
#include <boost/array.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/format.hpp>
#include <cmath>
#include <climits>
#include <sstream>
#include <fstream>

#include <csignal>
#include <iostream>
#include <string>
#include <utility>
#include <vector>
#include "common.h"
#include <QProcess>
#include <QString>

class UIRepository;

class Axis {
public:
	// 軸のラベル
	std::string label;
	// 軸のインデックスに乗ずる数字
	double weight;
	// この軸の次元の要素インデックス、固定値でない場合は特殊なマクロ定数が入る。
	int index;
	// 各軸の値の表示範囲(インデックスでは無くて重みを掛けた実際の値)
	std::pair<std::string ,std::string> range;
	Axis() {
		weight = 1;
		index = 0;
	}
	~Axis() {}
};

template <typename T1=int>
class ScatterPlot {
	std::vector<std::vector<T1> > *lists;
	std::string title;
	int plot_style; // PlotStyle
	std::string x_label, y_label, z_label;
public:
	// gnuplotの表示マクロ
	class PlotStyle {public: enum {UNDEFINE=-1, DOTS, POINTS, LINES, LINESPOINTS};};

	ScatterPlot(std::vector<std::vector<T1> > *lists) {
		// std::vector[][3]; ２次元目の深さは３であるとする。
		for (uint i=0; i < lists->size(); i++) {
			if ((*lists)[i].size() != 3) {
				THROW_CRITICAL("Plot coordinate size is not 3.");
			}
		}
		this->lists = lists;
		x_label = "X";
		y_label = "Y";
		z_label = "Z";
	}
	~ScatterPlot() {}

	void setStyle(int plot_style) {
		this->plot_style = plot_style;
	}
	void setTitle(std::string title) {
		this->title = title;
	}
	void setLabels(std::string x, std::string y, std::string z) {
		x_label = x;
		y_label = y;
		z_label = z;
	}
	std::string plot(std::string fullpath) {
		std::string file = createFile(fullpath);
		std::string command;
//		command = std::string("gnuplot -persist ") + file + " &";
//		int ret = system(command.c_str());
//		(void)ret;
		qint64 pid;
		int ret = QProcess::startDetached("gnuplot", QStringList() << "-persist" << QString(file.c_str()), QDir::currentPath(), &pid);
		UIRepository* repos = UIRepository::getRepository();
		if (!ret) {
			THROW_CRITICAL(QString("Command error: gnuplot --persist %1").arg(file.c_str()));
		}
		repos->addPid(pid);
		repos->addTempFile(file);
		return file;
	}
	std::string createFile(std::string fullpath) {
		FILE *fp;
		std::ostringstream tmp_os;

		// テンポラリファイルオープン
		char filename[256];
		sprintf(filename, "%s/temp-XXXXXX", fullpath.c_str());

		int fd =  mkstemp(filename);
		fp = fdopen(fd, "w");
		if (!fp) {
			THROW_WARNING(std::string("Can't open tempfile :") + filename);
		}
		fp = fopen(filename, "w");
		// タイトル
		fprintf(fp, "set title '%s'\n", title.c_str());
		// 表示範囲
/*
		fprintf(fp,"set xrange [%s:%s]\n", axises[x_index].range.first.c_str(), axises[x_index].range.second.c_str());
		fprintf(fp,"set yrange [%s:%s]\n", axises[y_index].range.first.c_str(), axises[y_index].range.second.c_str());
		fprintf(fp,"set zrange [%s:%s]\n", axises[z_index].range.first.c_str(), axises[z_index].range.second.c_str());
*/
		// 軸のタイトル
		fprintf(fp,"set xlabel '%s'\n", x_label.c_str());
		fprintf(fp,"set ylabel '%s'\n", y_label.c_str());
		fprintf(fp,"set zlabel '%s'\n", z_label.c_str());

		// XYZ軸
		fprintf(fp,"set xzeroaxis lt -1\n");
		fprintf(fp,"set yzeroaxis lt -1\n");
		fprintf(fp,"set zeroaxis lt -1\n");
		fprintf(fp,"set ticslevel 0\n"); // Z軸交点固定

		// プロットスタイル
		{
			std::string style_string;
			switch(plot_style) {
				case PlotStyle::UNDEFINE:
					style_string = "";
					break;
				case PlotStyle::DOTS:
					style_string = "with dots";
					break;
				case PlotStyle::POINTS:
					style_string = "with points";
					break;
				case PlotStyle::LINES:
					style_string = "with lines";
					break;
				case PlotStyle::LINESPOINTS:
					style_string = "with linespoints";
				default:
					break;
			}
			fprintf(fp, "splot '-' notitle %s\n", style_string.c_str());
		}
		// 内容の表示
		for (uint i = 0; i < lists->size(); i++) {
			T1 x = (*lists)[i][0];
			T1 y = (*lists)[i][1];
			T1 z = (*lists)[i][2];
			std::ostringstream tmp_os;
			tmp_os << x << " " << y << " " << z;
			fprintf(fp, "%s\n", tmp_os.str().c_str());
			tmp_os.str("");
		}
		fprintf(fp, "e\n");
		fprintf(fp, "pause -1\n");
		fclose(fp);

		return std::string(filename);
	}

};

template <typename T1=int, int T2=3>
class GNUPlot {
public:
	// gnuplotの表示マクロ
	class PlotStyle {public: enum {UNDEFINE=-1, DOTS, POINTS, LINES, LINESPOINTS};};
	class ColorMap  {public: enum {UNDEFINE=-1, PM3D, PM3D_MAP, PM3D_AT_B};};
	class Surface   {public: enum {UNDEFINE=-1, LINE, MESH};};

	// 軸を表す表示マクロ
	class AXIS {public: enum {X=INT_MAX-2, Y, Z};};

	int plot_style;
	int color_map;
	int surface;

	// Z軸の単位
	class ZUnit   {public: enum {RAW, REAL, IMAG, AMPLITUDE, DB, PHASE};}; // RAW以外は複素数の対応
	int z_unit;

	// 配列データ
	boost::multi_array<T1, T2>* matrix;

	// グラフのタイトル関連
	std::string title;

	// マトリックスの表示対象となる次元のインデックス
	// 必ずZ軸に配列要素を表示する仕様なので、z_indexは [マトリックスの次元数 +1]の固定値である。
	int x_index;
	int y_index;
	int z_index;

	// Z軸の情報も含まれているので、matrixの次元数よりも＋１大きい
	// 添字がmatrixの各次元に対応している。最後の要素はZ軸の情報である。
	std::vector<Axis> axises;

	// このコンストラクタのあとに、必ずsetMatrix()で初期化してから使う事
	GNUPlot() {
		plot_style = PlotStyle::UNDEFINE;
		color_map = ColorMap::UNDEFINE;
		surface = Surface::UNDEFINE;
		z_unit = ZUnit::AMPLITUDE;
		matrix = NULL;
	}
	GNUPlot(boost::multi_array<T1, T2>* matrix, std::vector<int> indexes) {
		setMatrix(matrix, indexes);
		plot_style = PlotStyle::UNDEFINE;
		color_map = ColorMap::UNDEFINE;
		surface = Surface::UNDEFINE;
		z_unit = ZUnit::AMPLITUDE;
	}
	void setMatrix(boost::multi_array<T1, T2>* matrix, std::vector<int> indexes) {
		// メインデータの保持
		this->matrix = matrix;
		if (!matrix->num_dimensions() == indexes.size() && !matrix->num_dimensions() == indexes.size() + 1) {
			// "マトリックスの次元とインデックスの次元が異なります。(インデックスの次元数は、マトリックスの次元数+1も有効です)"
			THROW_CRITICAL("Incompatible matrix's dimension and index's dimension");
		}
		// axises のインデックスの初期化
		axises.resize(matrix->num_dimensions() + 1); // +1 はZ軸
		this->setIndexes(indexes);
	}
	void setStyle(int plot_style, int surface, int color_map) {
		this->plot_style = plot_style;
		this->color_map = color_map;
		this->surface = surface;
	}
	~GNUPlot() {
	}
	void setTitle(std::string title) {
		this->title = title;
	}
	// 引数の次元は、マトリックスの次元に対応しなければならない
	void setLabels(std::vector<std::string> labels) {
		for (uint i=0; i< labels.size() && i < matrix->num_dimensions() + 1; i++) {
			axises[i].label = labels[i];
		}
	}
	// マトリックスの次元とは関係ない
	void setCurrentLabels(std::string x, std::string y, std::string z) {
		axises[x_index].label = x;
		axises[y_index].label = y;
		axises[z_index].label = z;
	}
	// 引数の次元は、マトリックスの次元に対応しなければならない
	void setWeights(std::vector<double> weights) {
		for (uint i=0; i< weights.size() && i < matrix->num_dimensions() + 1; i++) {
			axises[i].weight = weights[i];
		}
	}
	// マトリックスの次元とは関係ない
	void setCurrentWeight(double x, double y, double z) {
		axises[x_index].weight = x;
		axises[y_index].weight = y;
		axises[z_index].weight = z;
	}
	// 引数の次元は、マトリックスの次元に対応しなければならない
	void setRanges(std::vector<std::pair<std::string, std::string> > ranges) {
		for (uint i=0; i<ranges.size() && i < matrix->num_dimensions() + 1; i++) {
			axises[i].range = ranges[i];
		}
	}
	// マトリックスの次元とは関係ない
	void setCurrentRange(std::pair<std::string, std::string> x, std::pair<std::string, std::string> y, std::pair<std::string, std::string> z) {
		axises[x_index].range = x;
		axises[y_index].range = y;
		axises[z_index].range = z;
	}
	void setIndexes(std::vector<int> indexes) {
		bool find_x=false, find_y=false, find_z=false;
		for (uint i=0; i < indexes.size(); i++) {
			//　軸のインデックスのコピー
			axises[i].index = indexes[i];
			// x軸、y軸の存在チェック
			if (indexes[i] == GNUPlot<>::AXIS::X) {
				x_index = i;
				find_x = true;
			} else if (indexes[i] == GNUPlot<>::AXIS::Y) {
				y_index = i;
				find_y = true;
			} else if (indexes[i] == GNUPlot<>::AXIS::Z) { // Z軸が定義されている場合
				z_index = i;
				find_z = true;
			}
		}
		if (find_x == false || find_y == false || (find_z && (uint)z_index != matrix->num_dimensions())) {
			if (find_x == false || find_y == false ) {
				// "X,Y軸のインデックスが指定されていません"
				THROW_CRITICAL("Axis's index is Undefined on X or Y");
			} else {
				// "Z軸のインデックスが指定されていますが、無効なインデックスです。"
				THROW_CRITICAL("Z axis's Index is defined, but it is inappropriate");
			}
		}
		if (matrix->num_dimensions() == indexes.size()) { // Z軸が未定義の場合
			z_index = matrix->num_dimensions();
			axises[z_index].index = AXIS::Z;
		}
	}
	void setZUnit(int value) {
		z_unit = value;
	}
	std::string plot(std::string fullpath) {
		std::string file = plotToTempFile(fullpath);
		// std::string command;
		// command = std::string("gnuplot -persist ") + file + " &";
		// int ret = system(command.c_str());
		// (void)ret;
		qint64 pid;
		int ret = QProcess::startDetached("gnuplot", QStringList() << "-persist" << QString(file.c_str()), QDir::currentPath(), &pid);
		UIRepository* repos = UIRepository::getRepository();
		if (!ret) {
			THROW_CRITICAL(QString("Command error: gnuplot --persist %1").arg(file.c_str()));
		}
		repos->addPid(pid);
		repos->addTempFile(file);


		return file;
	}
	std::string plotToTempFile(std::string pathname) {

		// テンポラリファイル
		char fname[256];
		sprintf(fname, "%s/temp-XXXXXX", pathname.c_str());
		int ret = mkstemp(fname);
		(void)ret;
		std::ofstream ofs(fname);
		if(ofs.fail())
		{
			THROW_WARNING(std::string("Can't open file:") + fname);
		}
		plotToStream(ofs);

		return fname;
	}
	void plotToFile(std::string fname) {

		std::ofstream ofs(fname.c_str());
		if(ofs.fail())
		{
			THROW_WARNING(std::string("Can't open file:") + fname);
		}
		plotToStream(ofs);

	}
	void exportCSVToFile(std::string fname, bool complex=true) {
		std::ofstream ofs(fname.c_str());
		if(ofs.fail())
		{
			THROW_WARNING(std::string("Can't open file:") + fname);
		}
		exportCSVToStream(ofs, complex);
	}

	/*
	 * X軸方向をテキストの行,Y軸方向をテキストの列に対応するようにして、CSV形式でそのまま出力します
	 * 各軸のWeightは考慮されません
	 * complexがfalseの場合は、データが複素数でも実部のみ出力します
	 */
	void exportCSVToStream(std::ostream& os, bool complex=true) {

		boost::array<int, T2> index;
		// プロットしない固定次元をindex配列に指定する
		for (uint i=0; i < axises.size() - 1; i++) {
			index[i] = axises[i].index;
		}

		// プロットする２次元分の処理
		for (uint y = 0; y < matrix->shape()[y_index]; y++) {
			for (uint x = 0; x < matrix->shape()[x_index]; x++) {
				index[x_index] = x;
				index[y_index] = y;
				if (x) {
					os <<  ",";
				}
				os  << toString((*matrix)(index), complex);
			}
			os <<  "\n";
		}
	}

	void plotToStream(std::ostream& os) {

		// タイトル
		os << boost::format("set title '%s'\n") %title;

		// 表示範囲
		os << boost::format("set xrange [%s:%s]\n") %axises[x_index].range.first %axises[x_index].range.second;
		os << boost::format("set yrange [%s:%s]\n") %axises[y_index].range.first %axises[y_index].range.second;
		os << boost::format("set zrange [%s:%s]\n") %axises[z_index].range.first %axises[z_index].range.second;
		// 軸のタイトル
		os << boost::format("set xlabel '%s'\n") %axises[x_index].label;
		os << boost::format("set ylabel '%s'\n") %axises[y_index].label;
		os << boost::format("set zlabel '%s'\n") %axises[z_index].label;

		// XYZ軸
		os << "set xzeroaxis lt -1\n";
		os << "set yzeroaxis lt -1\n";
		os << "set zeroaxis lt -1\n";
		os << "set ticslevel 0\n"; // Z軸交点固定

		// カラーマップ
		{
			std::string color_map_string;
			switch(color_map) {
				case ColorMap::PM3D:
					color_map_string = "pm3d";
					break;
				case ColorMap::PM3D_MAP:
					color_map_string = "pm3d map";
					break;
				case ColorMap::PM3D_AT_B:
					color_map_string = "pm3d at b";
					break;
				default:
					color_map_string = "";
					break;
			}
			if (color_map_string != "")
				os <<  boost::format("set %s\n") %color_map_string;
		}
		// プロットスタイル
		{
			std::string style_string;
			switch(plot_style) {
				case PlotStyle::UNDEFINE:
					if (color_map == ColorMap::PM3D) {
						style_string = "with pm3d";
					} else {
						style_string = "";
					}
					break;
				case PlotStyle::DOTS:
					style_string = "with dots";
					break;
				case PlotStyle::POINTS:
					style_string = "with points";
					break;
				case PlotStyle::LINES:
					style_string = "with lines";
					break;
				case PlotStyle::LINESPOINTS:
					style_string = "with linespoints";
			}
			os << boost::format("splot '-' notitle %s\n") %style_string;
		}

		boost::array<int, T2> index;
		// プロットしない固定次元をindex配列に指定する
		for (uint i=0; i < axises.size() - 1; i++) {
			index[i] = axises[i].index;
		}

		// プロットする２次元分の処理
		for (uint y = 0; y < matrix->shape()[y_index]; y++) { // このループはlines の時に x-z平面方向の線を引くようにしている。
			for (uint x = 0; x < matrix->shape()[x_index]; x++) {
				index[x_index] = x;
				index[y_index] = y;
				os << boost::format("%f %f %f\n") %(x * axises[x_index].weight) %(y * axises[y_index].weight) %(calcZValue(z_unit, (*matrix)(index)) *axises[z_index].weight);
			}
			// Surface
			{
				std::string surface_string;
				switch(surface) {
					case Surface::MESH:
						surface_string = "";
						break;
					case Surface::LINE:
						surface_string = "\n";
						break;
					default:
						surface_string = "";
						break;
				}
				os <<  boost::format("%s\n") %surface_string;
			}
		}
		os <<  "e\n";
		os <<  "pause -1\n";

		return;
	}
};

/*
 * valueの数字をz_unitの単位に変換します。
 * 複素数の場合はz_unitの単位に変換されるので、かならず実数に変換されます。
 */
template <typename T1> double calcZValue(int z_unit, T1 value) {

	switch(z_unit) {
		case GNUPlot<>::ZUnit::RAW:
			return value;
			break;
		default:
			//"Z軸値の値を実数に変換出来ません"
			THROW_CRITICAL("Can't conversion Z'value and real number");
			break;
	}
}
/*
 * valueの数字を文字列に変換します。
 * 複素数の場合は 実部+虚部i の形式に変換します。
 * complexがfalseの場合は実部のみに変換します。
 */
template <typename T1> std::string toString(T1 value, bool complex=true) {
(void)complex;
	std::string str = (boost::format("%1%") % value).str();
	return str;
}

template <> double calcZValue<std::complex<double> >(int z_unit, std::complex<double> value);


namespace HT {
class UIApp;

namespace MIC {
	class UIDocument;
}}
namespace HT {
namespace GSS {
	class UIDocument;
} //namespace GSS
namespace MUSIC {
	class UIDocument;
} //namespace MUSIC
namespace MIC {
	class UIDocument;
} //namespace MIC

} //namespace
#endif
