#ifndef UITFDATA_H
#define UITFDATA_H

#include <iostream>
#include <string>
#include <sstream>
#include <cmath>
#include <vector>
#include <boost/shared_ptr.hpp>
#include <complex>
#include "UIApp.h"
#include "UIData.h"
#include "UITFDocument.h"
#include "UIGraph.h"
#include "Matrix.h"
#include "libharkio/libharkio.h"

//! GSSのTFファイルのバイナリ中のキーとなる数字
//{
const int NAE_TFFILE_FILEINFO_ID	= 0x00310000;
const int NAE_TFFILE_SRCPOS_ID		= 0x00320000;
const int NAE_TFFILE_MICPOS_ID		= 0x00330000;
const int NAE_TFFILE_FREQLINE_ID	= 0x00340000;
const int NAE_TFFILE_TFVALUE_ID	= 0x00350000;
//}

// GSS_TFData.cに含まれているものと同じデータ型。ファイル読み書き時に利用している
//\{
typedef struct {
	double	dX;
	double	dY;
	double	dZ;
	double	dTheta;
	double	dPhi;
	long	lFlag;
	char	cPad[20];
} NAE_TFFILE_COORD;

typedef struct {
	double dRe;
	double dIm;
} NAE_TFFILE_COMPLEX;
//\}

namespace HT {
class UIDocument;
namespace GSS {

typedef float HT_FLT_DATA_TYPE; //! インパルス応答ファイルのデータ型

//! SRCファイル各行の情報を保存するクラス
template <typename T>
class UISrc {
public:
	//! 座標タイプ
	/*!
	 * この値は、このクラスで保持してないが、micarrayシステムの構造体メンバを判別する為に
	 * 使用するので、ここに宣言している。
	 */
	typedef enum COORDS{CARTESIAN=4+2+1, POLAR=16+8} COORDS;
	typedef boost::shared_ptr<UISrc<T> > shared_ptr;
	T x;
	T y;
	T z;
	//! 新規にTFファイルを作成する時だけに使用されるデータ
	//\{
	std::string implement_file;	//!srcinfテキストに埋め込まれたfltファイル名
	int cut_start;			//! srcinfファイルのfltファイル名の後ろの数
	//\}
	//!デバックプリント関数
	void print() {
		std::cout << x << "," << y << "," << z << "," << implement_file << "," << cut_start << std::endl;
	}

};

// メンバ関数用パラメータ
class CreateTFParam {
public:
	std::string mic_file;
	std::string src_file;
	std::string mic_string;
	std::string src_string;

	int NFFT;
	int NFS;
	int IMPULSE_RESPONSE_LEN;
};

class moveToHARKIOParam {
public:
	std::vector<HT::UISrc<HT_DATA_TYPE> >* src_list;
	std::vector<HT::UIMic<HT_DATA_TYPE> >* mic_list;
	TFMatrix<HT_COMPLEX_DATA_TYPE>* tf_matrix;
	int NMic;
	int NFFT;
	int NIR; // インパルスレスポンス長
	int NFS;
};

//!  GSSのＴＦファイルのデータコンテナクラス
class UIData : public HT::UIData{
	bool tsp_canceled;
private:
	// 旧形式データ
	boost::shared_ptr<TFMatrix<HT_COMPLEX_DATA_TYPE> > matrix; //! 伝達関数マトリックス
	boost::shared_ptr<std::vector<HT::UIMic<HT_DATA_TYPE> > > mic; //! マイク座標配列
	boost::shared_ptr<std::vector<HT::UISrc<HT_DATA_TYPE> > > src; //! 音源位置座標配列
	int version; //! バージョン
	int file_type;//! FileType(SM,TF)

	// メッシュデータ情報
	bool use_mesh;
	int mesh_x;
	int mesh_y;
	int mesh_z;

	// 正規化方法
	int normalization;

	int NFreq;	//! 周波数ライン数
	int NSrc;	//! Src数
	int NMic;	//! マイク数
	int FFT;	//! FFT長
	int FS;		//! サンプリング周波数

	// 新形式 harkio データ
	boost::shared_ptr<harkio::HARK> harkio;
public:
	void setHARKIO(harkio::HARK* harkio) {
		if (this->harkio.get() != harkio) {
			this->harkio.reset(harkio);
		}
	}
	harkio::HARK* getHARKIO() {
		return harkio.get();
	}
	bool isHARKIO() {
		if (harkio.get()) {
			return true;
		} else {
			return false;
		}
	}

	UIDocument* getDocument();
	harkio::HARK* moveToHARKIO(const moveToHARKIOParam& param);

	//! バージョン
	typedef enum {V_ZERO=0,V_ONE=1} VERSION;
	//! ファイルタイプ
	typedef enum {TF_CONJ=0,SM=1,AUTO=2} FILE_TYPE;

	//! 正規化種類
#if 0
	typedef enum {NEW,OLD} NORMALIZATION;
#else // 2012-02-22 (s.noguchi)正規化の処理方法を追加
	typedef enum {CH,CH_FRQ,CH_FRQ_DIR,NONE} NORMALIZATION;
#endif

	//! ファイルタイプのシンボル配列

	// 配列の添え字はFILE_TYPEと同じ
	static const char *file_type_name[3];
	UIData(UIDocument *document);
	UIData();
	~UIData();

	//! 伝達関数マトリックスの各次元を表すシンボル
	typedef enum {
		 FRQ = 0
		,MIC = 1
		,SRC = 2
	} MATRIX_SUBSCRIPT;
	//! 伝達関数マトリックス全体の次元数
	typedef enum {DM3 = 3} MATRIX_DIM;

	void setNFreq(int value) {
		NFreq = value;
	}
	int getNFreq() {
		return NFreq;
	}

	void setNSrc(int value) {
		NSrc = value;
	}
	int getNSrc() {
		return NSrc;
	}

	void setNMic(int value) {
		NMic = value;
	}
	int getNMic() {
		return NMic;
	}

	void setNFFT(int value) {
		FFT = value;
	}
	int getNFFT() {
		return FFT;
	}

	void setNFS(int value) {
		FS = value;
	}
	int getNFS() {
		return FS;
	}


	void setUseMesh(bool value) {
		use_mesh = value;
	}
	bool getUseMesh() {
		return use_mesh;
	}

	void setVersion(int value) {
		version = value;
	}
	int getVersion() {
		return version;
	}

	void setFileType(int value) {
		file_type = value;
	}
	int getFileType() {
		return file_type;
	}

	void setMeshFlg(bool value) {
		use_mesh= value;
	}
	bool getMeshFlg() {
		return use_mesh;
	}

	void setMesh(int x, int y, int z) {
		mesh_x =x;
		mesh_y =y;
		mesh_z =z;
	}
	void getMesh(int *x, int *y, int *z) {
		*x = mesh_x;
		*y = mesh_y;
		*z = mesh_z;
	}

	void setNormalization(int normal) {
		normalization = normal;
	}
	int getNormalization() {
		return normalization;
	}

	//! 伝達関数マトリックス Matrix
	template<typename T> void setTFMatrix(T *matrix) {
		this->matrix.reset(matrix);
	}
	TFMatrix<HT_COMPLEX_DATA_TYPE> *getMatrix() {
		return matrix.get();
	}

	//! マイク座標配列 UIMic
	template<typename T> void setMic(T *mic) {
		this->mic.reset(mic);
	}
	std::vector<HT::UIMic<HT_DATA_TYPE> > *getMic() {
		return mic.get();
	}

	//! 音源座標配列 Src
	template<typename T> void setSrc(T *src) {
		this->src.reset(src);
		return;
	}
	std::vector<HT::UISrc<HT_DATA_TYPE> > *getSrc() {
		return src.get();
	}

	//! 伝達関数の新規作成関数
	int createTFData(const CreateTFParam& param);
	//! TFファイルとしてデータを保存する関数
	int saveTFFile(std::string fullpath_file);
	//! 逆変換時のサブ関数
	TFMatrix<HT_COMPLEX_DATA_TYPE>* IFFT(TFMatrix<HT_COMPLEX_DATA_TYPE>& matrix, int FFT, int NMIC, int NSRC);
	void setTSPCanceled(bool value) {
		tsp_canceled = value;
	}
public:
};

//! ファイルタイプのシンボル配列を初期化
__declspec(selectany) const char *UIData::file_type_name[3] = {"TF_CONJ", "SM","AUTO"};

} //namespace GSS
} //namespace
#endif
