#include "UITFDocument.h"
#include "UIGraph.h"
#include "UITFData.h"
#include "Matrix.h"
#include "common.h"
#include "libharkio/libharkio.h"
#include <QString>

using namespace std;
using namespace boost;
using namespace harkio;

namespace HT {
namespace GSS {

UIDocument::UIDocument(std::string name, std::string filename)
{
	setName(name);
	setFileName(filename);
};

UIDocument::UIDocument() {}
UIDocument::~UIDocument() {}

UIData *UIDocument::newTFData(UIDocument *doc) {
	data.reset(new UIData(doc));
	return data.get();
}
UIData* UIDocument::getUIData(bool exception) {
	if (exception && !data) {
		THROW_CRITICAL("Data is null");
	}
	return data.get();
}
HT::UIData* UIDocument::getBaseUIData() {
	return data.get();
}

void UIDocument::loadFileByLib(string filename) {

	try {
		UIData *data = newTFData(this);
		setFileName(filename);
		// ライブラリを使った読み込み
		HARK *hark = new HARK;
		data->setHARKIO(hark);
		hark->readFile(filename);
	} catch (harkio::Error &e) {
		data->setHARKIO(NULL);
		THROW_CRITICAL(e.getMsg().c_str());
	}
}

void UIDocument::loadFileNormally(string filename) {
IFDEBUG(DP("loadFileNormally"));

	FILE *fp = NULL;
	HT::GSS::UIData *data = NULL;
	vector<HT::UISrc<HT_DATA_TYPE> > *src = NULL;
	vector<UIMic<HT_DATA_TYPE> > *mic = NULL;
	TFMatrix<HT_COMPLEX_DATA_TYPE> *tf_matrix = NULL;
	NAE_TFFILE_COMPLEX *gss_tf = NULL;
	setFileName(filename);
	try {

		int int_buf;
		int read_size;
		// ファイルオープン
		fp = fopen(filename.c_str(), "r");
		if (fp == NULL) {
			THROW_WARNING(QString("Can't open file :%1").arg(filename.c_str()));
		}
		// HEADER
		IFDEBUG(DP("HEADER"));
		// ID
		read_size = fread(&int_buf, sizeof(int), 1, fp);
		if (read_size != 1 || int_buf != NAE_TFFILE_FILEINFO_ID) {
			THROW_CRITICAL("Can't read NAE_TFFILE_FILEINFO_ID.");
		}
		// SIZE
		read_size = fread(&int_buf, sizeof(int), 1, fp);
		if (read_size != 1) {
			THROW_CRITICAL("Can't read NAE_TFDBFileHeader Byte size.");
		}
		// 一応、ヘッダーの構造体サイズと比較してByteSizeチェック
		// 異なっていても続行できるようにする
		// 昔のファイルはここでエラーが発生することが多いので無視する
/*
		if (int_buf != sizeof(NAE_TFDBFileHeader)) {
			//		THROW_WARNING("NAE_TFDBFileHeader size is incorrect.");
			string message = "Read value " + boost::lexical_cast<string>(int_buf)
					+ " is different from expected value which is NAE_TFDBFileHeader size " + boost::lexical_cast<string>(
					sizeof(NAE_TFDBFileHeader)) + ".\n" + "Is reading continued ?";
		}
*/
		// 上でSIZEを読んでいるが、特に使用しない
		// Header 構造体へ読み込み
		NAE_TFDBFileHeader header;
		memset(&header, 0, sizeof(NAE_TFDBFileHeader));
		read_size = fread(&header, sizeof(NAE_TFDBFileHeader), 1, fp);
		if (read_size != 1) {
			THROW_CRITICAL("Can't read NAE_TFDBFileHeader data.");
		}
		// Headerを UIDataインスタンスへコピー
		data = newTFData(this);
		data->setNFFT(header.lNFFT); // FFT長
		data->setNFS(header.lFs); // サンプリング周波数
		data->setMeshFlg(header.lMeshFlg); // メッシュの使用フラグ
		data->setMesh(header.lMeshX, header.lMeshY, header.lMeshZ); // メッシュ格子数
		data->setVersion(header.iVersion); // バージョン番号
		data->setFileType(header.iFileType); // ファイルタイプ
		data->setNFreq(header.lNFreq); // 周波数ライン数
		data->setNSrc(header.lNSrc); // ソース数
		data->setNMic(header.lNMic); // マイク数

		// SRC
		IFDEBUG(DP("SRC"));
		// ID
		read_size = fread(&int_buf, sizeof(int), 1, fp);
		if (read_size != 1 || int_buf != NAE_TFFILE_SRCPOS_ID) {
			THROW_CRITICAL("Can't read NAE_TFFILE_SRCPOS_ID.");
		}
		// SIZE
		read_size = fread(&int_buf, sizeof(int), 1, fp);
		if (read_size != 1) {
			THROW_CRITICAL("Can't read NAE_TFFILE_COORD(Src) Byte size.");
		}
		// 一応、ヘッダーのSRC数と比較してByteSizeチェック
		if ((unsigned) int_buf != sizeof(NAE_TFFILE_COORD) * data->getNSrc()) {
			THROW_CRITICAL("NAE_TFFILE_COORD * NSrc is incorrect.");
		}
		// 上でSIZEを読んでいるが、特に使用しない
		// UISrc データ読み込み
		NAE_TFFILE_COORD gss_src[sizeof(NAE_TFFILE_COORD) * (data->getNSrc())];
		memset(&gss_src, 0, sizeof(NAE_TFFILE_COORD) * (data->getNSrc()));
		read_size = fread(&gss_src, sizeof(NAE_TFFILE_COORD), data->getNSrc(), fp);
		if (read_size != data->getNSrc()) {
			THROW_CRITICAL("Can't read all of Src position.");
		}
		// UISrc インスタンスへコピー
		src = new vector<HT::UISrc<HT_DATA_TYPE> > ();
		src->resize(data->getNSrc());
		data->setSrc(src); // UISrc インスタンスポインタ保持
		for (int i = 0; i < data->getNSrc(); i++) {
			// ビットフラグが共に0の時
			int flag = gss_src[i].lFlag;
			if (!(flag & (UISrc<HT_DATA_TYPE>::CARTESIAN)) && !(flag & (UISrc<HT_DATA_TYPE>::POLAR))) {
				THROW_CRITICAL(QString("Enable flag(%1) is neither Cartesian or Polar coordinate Src:%2").arg(flag).arg(i));
			}
			// 両方の座標系にビットフラグが立っている時
			if ((flag & (UISrc<HT_DATA_TYPE>::CARTESIAN)) && (flag & (UISrc<HT_DATA_TYPE>::POLAR))) {
				THROW_CRITICAL(QString("Enable flag(%1). Both Cartesian and Polar coordinate flag is enable. Src:%2").arg(flag).arg(i));
			}

			// UISrcインスタンスはデカルト座標に統一して保持する
			if (flag & (UISrc<HT_DATA_TYPE>::CARTESIAN)) {
				(*src)[i].x = gss_src[i].dX;
				(*src)[i].y = gss_src[i].dY;
				(*src)[i].z = gss_src[i].dZ;

			} else if (flag & (UISrc<HT_DATA_TYPE>::POLAR)) {
				vector<double> coords = getPolarToCartesian(1.0, gss_src[i].dTheta, gss_src[i].dPhi);
				(*src)[i].x = coords[0];
				(*src)[i].y = coords[1];
				(*src)[i].z = coords[2];
			}
		}

		// MIC
		IFDEBUG(DP("MIC"));
		// ID
		read_size = fread(&int_buf, sizeof(int), 1, fp);
		if (read_size != 1 || int_buf != NAE_TFFILE_MICPOS_ID) {
			THROW_CRITICAL("Can't read NAE_TFFILE_MICPOS_ID.");
		}
		// SIZE
		read_size = fread(&int_buf, sizeof(int), 1, fp);
		if (read_size != 1) {
			THROW_CRITICAL("Can't read NAE_TFFILE_COORD(UIMic) Byte size.");
		}
		// 一応、ヘッダーのMIC数と比較してByteSizeチェック
		if ((unsigned) int_buf != sizeof(NAE_TFFILE_COORD) * data->getNMic()) {
			THROW_CRITICAL("NAE_TFFILE_COORD * NMic is incorrect.");
		}
		// 上でSIZEを読んでいるが、特に使用しない
		// UIMic データ読み込み
		NAE_TFFILE_COORD gss_mic[sizeof(NAE_TFFILE_COORD) * (data->getNMic())];
		memset(&gss_mic, 0, sizeof(NAE_TFFILE_COORD) * (data->getNSrc()));
		read_size = fread(&gss_mic, sizeof(NAE_TFFILE_COORD), data->getNMic(), fp);
		if (read_size != data->getNMic()) {
			THROW_CRITICAL("Can't read all of Mic position.");
		}
		// UIMic インスタンスへコピー
		mic = new vector<UIMic<HT_DATA_TYPE> > ();
		mic->resize(data->getNMic());
		data->setMic(mic); // UIMic インスタンスポインタ保持
		for (int i = 0; i < data->getNMic(); i++) {
			// 座標系チェック

			// ビットフラグが共に0の時
			int flag = gss_mic[i].lFlag;
			if (!(flag & (UIMic<HT_DATA_TYPE>::CARTESIAN)) && !(flag & (UIMic<HT_DATA_TYPE>::POLAR))) {
				THROW_CRITICAL(QString("Enable flag(%1) is neither Cartesian and Polar coordinate flag is enable. Mic:%2").arg(flag).arg(i));
			}
			// 両方の座標系にビットフラグが立っている時
			if ((flag & (UIMic<HT_DATA_TYPE>::CARTESIAN)) && (flag & (UIMic<HT_DATA_TYPE>::POLAR))) {
				THROW_CRITICAL(QString("Enable flag(%1). Both Cartesian and Polar coordinate flag is enable. Mic:%2").arg(flag).arg(i));
			}
			// UIMicインスタンスへにはデカルト座標に統一して保持する
			if (flag & (UIMic<HT_DATA_TYPE>::CARTESIAN)) {
				(*mic)[i].x = gss_mic[i].dX;
				(*mic)[i].y = gss_mic[i].dY;
				(*mic)[i].z = gss_mic[i].dZ;

			} else if (flag & (UIMic<HT_DATA_TYPE>::POLAR)) {
				vector<double> coords = getPolarToCartesian(1.0, gss_mic[i].dTheta, gss_mic[i].dPhi);
				(*mic)[i].x = coords[0];
				(*mic)[i].y = coords[1];
				(*mic)[i].z = coords[2];
			}
		}

		// FreqLine
		IFDEBUG(DP("FreqLine"));
		read_size = fread(&int_buf, sizeof(int), 1, fp);
		if (read_size != 1 || int_buf != NAE_TFFILE_FREQLINE_ID) {
			THROW_CRITICAL("Can't read NAE_TFFILE_FREQLINE_ID.");
		}
		// SIZE
		read_size = fread(&int_buf, sizeof(int), 1, fp);
		if (read_size != 1) {
			THROW_CRITICAL("Can't read FreqLine Byte size.");
		}
		// 上でSIZEを読んでいるが、特に使用しない
		// 一応、FFTと比較してByteSizeチェック
		if ((unsigned) int_buf != sizeof(int) * (data->getNFFT() / 2 + 1)) {
			THROW_CRITICAL("FreqLine size is incorrect.");
		}
		// FreqLine データ読み込み
		int gss_freq[sizeof(int) * (data->getNFFT() / 2 + 1)];
		memset(&gss_freq, 0, sizeof(int) * (data->getNFFT() / 2 + 1));
		read_size = fread(&gss_freq, sizeof(int), data->getNFFT() / 2 + 1, fp);
		if (read_size != data->getNFFT() / 2 + 1) {
			THROW_CRITICAL("Can't read all of Freq line.");
		}
		// FreqLineに関して、特にデータの格納はない


		// TF Data
		IFDEBUG(DP("TFData"));
		read_size = fread(&int_buf, sizeof(int), 1, fp);
		if (read_size != 1 || int_buf != NAE_TFFILE_TFVALUE_ID) {
			THROW_CRITICAL("Can't read NAE_TFFILE_TFVALUE_ID.");
		}
		int size = (data->getNFFT() / 2 + 1) * data->getNMic() * data->getNSrc();
		// 注: TFDataに関しては、ここに、データサイズの指定は無い

		// TF Data データ読み込み
		gss_tf = (NAE_TFFILE_COMPLEX *) malloc(sizeof(NAE_TFFILE_COMPLEX) * size);
		memset(gss_tf, 0, sizeof(NAE_TFFILE_COMPLEX) * size);
		read_size = fread(gss_tf, sizeof(NAE_TFFILE_COMPLEX), size, fp);
		if (read_size != size) {
			THROW_CRITICAL("Can't read all of TF data.");
		}
		// 終端データの確認
		int end = ftell(fp);
		fseek(fp, 0, SEEK_END);
		if (ftell(fp) != end) {
			THROW_CRITICAL("Can't reach end of file.");
		}

		// Matrix インスタンスへコピー
		tf_matrix = new TFMatrix<HT_COMPLEX_DATA_TYPE> ();
		tf_matrix->resize(3, data->getNFFT() / 2 + 1, data->getNMic(), data->getNSrc());
		data->setTFMatrix(tf_matrix); // TF Matrixインスタンスポインタ保持

		int linear_offset = 0;
		// 分離行列の判定
		bool sm = false;
		if (data->getFileType() == HT::GSS::UIData::SM) {
			sm = true;
		}
		for (int nFFT = 0; nFFT < data->getNFreq(); nFFT++) {
			for (int nMic = 0; nMic < data->getNMic(); nMic++) {
				for (int nSrc = 0; nSrc < data->getNSrc(); nSrc++) {
					if (sm == false) {
						linear_offset = ((data->getNMic() * data->getNSrc()) * nFFT) + (data->getNMic() * nSrc) + (nMic);
					} else {
						// 分離行列の場合、読み込む順番が異なる
						linear_offset = (data->getNSrc() * (data->getNFFT() / 2 + 1) * nMic) + ((data->getNFFT() / 2 + 1) * nSrc) + (nFFT);
					}
					tf_matrix->matrix[nFFT][nMic][nSrc] = HT_COMPLEX_DATA_TYPE(gss_tf[linear_offset].dRe,
							gss_tf[linear_offset].dIm);
					// cout << nFFT << " "<< nMic << " "<< nSrc << " "<< tf_matrix->matrix[nFFT][nMic][nSrc] << endl;
				}
			}
		}
		free(gss_tf);

		IFDEBUG(DP("loadFile end"));

	} catch (...) {
		if (fp != NULL) {
			fclose(fp);
		}
		this->data.reset();
		throw;
	}
}

void UIDocument::loadFile(string filename) {
IFDEBUG(DP("# loadFile start"));

	// ファイル先頭の識別子から、Hharkio が扱えるかを判定する。
	HARKIO_RET ret = harkio::HARK::checkIdentifier(filename);
	if (ret == harkio::HARKIO_TRUE) {
		loadFileByLib(filename); // ライブラリ読み込み
	} else if (ret == harkio::HARKIO_FALSE) {
		loadFileNormally(filename); // 旧形式の読み込み
	} else {
		THROW_CRITICAL(harkio::Error::getInstance().getMsg().c_str());
	}
IFDEBUG(DP("# loadFile end"));
}



} //namespace GSS
} //namespace
