/*! \file common.h
 * 共通に使う、一般関数やデバック関数, typedef, define を定義したヘッダファイル
 */
#ifndef COMMON_H
#define COMMON_H

#include <stdio.h>
#include <stdlib.h>

#include <iostream>
#include <string>
#include <sstream>
#include <vector>
#include <map>
#include <boost/lexical_cast.hpp>

#include <QString>
#include <QDebug>
#include <QDir>

#include "UIException.h"

//#define DEBUG

//! \name VC++ Extension リンク時に、静的メンバ変数の重複を避ける
//\{
#define __declspec(x) __attribute__((x))
#define selectany weak
//\}

const std::string TSP_TO_IMPULSE = "calcSysImp";		// TSPからインパルス応答に変換するプログラム
const std::string HARK_TOOLS_DIR = "harktool3_utils";// 標準の外部プログラムのサブディレクトリ名
const std::string IMPULSE_ORG = "16384.little_endian.tsp";
const std::string TEMPORARY_PATH = ".harktool";

// QString -> char*の変換
#define QTOC(qstring) (qstring).toStdString().c_str()

//! \name DEBUG切り替えマクロ
//\{
#ifdef DEBUG
	#define IFDEBUG(statement) statement
#else
	#define IFDEBUG(statement)
#endif

#ifdef DEBUG
	#define IFNDEBUG(statement)
#else
	#define IFNDEBUG(statement) statement
#endif
//\}

#define CHOMPR(line) if(line.rfind("\r") != std::string::npos) {line.erase(line.rfind("\r"), 1);}
#define CHOMPN(line) if(line.rfind("\n") != std::string::npos) {line.erase(line.rfind("\n"), 1);}
#define CHOMP(line) CHOMPR(line) CHOMPN(line)

//! プリントデバック用のクラスとマクロ
/*!\name
 * 使用する際は、単にDP()マクロを呼び出すことでインスタンス化される
 * 使い方：DP("value=",value); の様な感じでプリントしたい物をカンマで区切って渡す
 */
//\{
class Debug{
public:
	static int call_count;
	QString file;
	int   line;
	Debug(QString file, int line) : file(file), line(line) {
	}
	template<typename T1> void operator ()() {
		std::cout << call_count ;
		call_count++;
	}
	template<typename T1> void operator ()(T1 arg1) {
		QString str;
		QDebug qd(&str);

		qd << file << line << arg1;
		qDebug() << str;
	}
	template<typename T1,typename T2> void operator ()(T1 arg1,T2 arg2) {
		QString str;
		QDebug qd(&str);
		qd << file << line;
		qd << arg1 << ", " << arg2 ;

		qDebug() << str;
	}
	template<typename T1,typename T2,typename T3> void operator ()(T1 arg1,T2 arg2, T3 arg3) {
		QString str;
		QDebug qd(&str);
		qd << file << line;

		qd << arg1 << ", " << arg2 << ", " << arg3 ;
		qDebug() << str;
	}
	template<typename T1,typename T2,typename T3,typename T4> void operator ()(T1 arg1, T2 arg2, T3 arg3, T4 arg4) {
		QString str;
		QDebug qd(&str);
		qd << file << line;

		qd << arg1 << ", " << arg2 << ", " << arg3 << ", " << arg4;

		qDebug() << str;
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5> void operator ()(T1 arg1, T2 arg2, T3 arg3, T4 arg4, T5 arg5) {
		QString str;
		QDebug qd(&str);
		qd << file << line;

		qd << arg1 << ", " << arg2 << ", " << arg3 << ", " << arg4 << ", " << arg5;

		qDebug() << str;
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6> void operator ()(T1 arg1, T2 arg2, T3 arg3, T4 arg4, T5 arg5, T6 arg6) {
		QString str;
		QDebug qd(&str);
		qd << file << line;

		qd << arg1 << ", " << arg2 << ", " << arg3 << ", " << arg4 << ", " << arg5 << ", " << arg6 ;

		qDebug() << str;
	}
};
#define DP Debug(__FILE__, __LINE__)
//\}


//! 座標変換関数 直交座標 -> 球座標
template<typename T> static std::vector<T> getCartesianToPolar(T x, T y, T z) {

	std::vector<T> result;
	result.resize(3);

	result[0] = sqrt(x*x + y*y + z*z);
	result[1] = atan2(sqrt(x*x + y*y), z);
	result[2] = atan2(y, x);
	return result; // 戻り値はradian
}

//! 座標変換関数 球座標 -> 直交座標
template<typename T> static std::vector<T> getPolarToCartesian(T r, T theta, T phi) {

	std::vector<T> result;
	result.resize(3);

	// 引数はradian
	result[0] = r * sin(theta)* cos(phi);
	result[1] = r * sin(theta)* sin(phi);
	result[2] = r * cos(theta);

	return result;
}


//! \name ファイル名とファイルパスを分割する関数
/*!
 * フルパスに限らず、相対パスでも問題はない
 */
//\{
//! ファイル名部分だけを取り出す
std::string fullPathToFile(std::string fullpath);
//! ファイルパス部分だけを取り出す
std::string fullPathToPath(std::string fullpath);
//\}

// harktoolの設定&テンポラリディレクトリの作成&取得
std::string harktoolDir();

// ユーザー名の取得
QString getUserName();

// 音声ファイルかどうかの確認だけを行う
bool checkSoundFile(std::string file);

class UIRepository {
private: //Singleton Pattern
	UIRepository() {}
private:
	std::vector<long> pids;
	std::vector<std::string> temp_file;
public:
	~UIRepository() {}
	static UIRepository *getRepository() {
		static UIRepository *repository=NULL;
		if (!repository) {
			repository = new UIRepository;
		}
		return repository;
	}
	void addPid(long pid) {
		pids.push_back(pid);
	}
	void addTempFile(std::string file) {
		temp_file.push_back(file);
	}
	void kill() {
		char command[128];
		for (uint i=0; i < pids.size(); i++) {
			sprintf(command, "kill %ld > /dev/null", pids[i]);
			int ret = system(command);
			(void)ret;
		}
	}
	void deleteTempFile() {
		char command[4098];
		for (uint i=0; i < temp_file.size(); i++) {
			sprintf(command, "rm -f %s > /dev/null", temp_file[i].c_str());
			int ret = system(command);
			(void)ret;
		}
	}
};


#if false
// パスからファイルを除いた、ディレクトリ名のみを取得する
// ファイルでない場合は何もしない
void pathOnly(QDir& dir);

// 相対パスファイル名(target)を、指定したディレクトリを基準とするパスに変更する
// targetが絶対パスの場合は何もしない
void setBasePath(QDir base, QDir& target);

// targetファイルを、baseが基準ディレクトリであるファイルに書き換える
QString basedPath(QString base, QString target);
#endif


#endif

